from autogluon.tabular import TabularDataset, TabularPredictor
from autogluon.core.metrics import BINARY_METRICS, MULTICLASS_METRICS
import pandas as pd
from typing import Tuple, Dict
from pathlib import Path
import numpy as np
import os
import h2o
from h2o.automl import H2OAutoML
from sklearn.metrics import accuracy_score, balanced_accuracy_score, precision_score, roc_auc_score, recall_score, f1_score,average_precision_score
from sklearn.preprocessing import LabelEncoder
import pickle
from abc import ABC, abstractmethod
from sklearn.model_selection import train_test_split
import logging
logging.basicConfig(level=logging.INFO, format="%(message)s")
YELLOW = "\033[93m"
RESET = "\033[0m"
BASE_DIR = Path(__file__).resolve().parents[2]
RES_SAVING_DIR = BASE_DIR/"converter_results"

class Converter(ABC):
    '''
    This is an abstract class defining all necessary methods, which have to be implemented in any converter class.
    As converter class uses different framework, tha parameters can be different in each class.
    '''
    @abstractmethod
    def create_predictions_dict(self)-> dict:
        '''Creates the dictionary with model names as keys and their class predictions vectors as values.
            Value vectors should be pandas.Series type for binary and multiclass task types. 
        '''
        pass
    @abstractmethod 
    def create_proba_predictions_dict(self)-> dict:
        '''Creates the dictionary with model names as keys and their probabilistic predictions as values.
            Values should be a pandas.DataFrame type for binary and multiclass task types.
        '''
        pass

    @abstractmethod
    def create_feature_importance_dict(self)->dict:
        '''Creates the dictionary with model names as keys and the list of features sorted by the feature importance as values'''
        pass

    @abstractmethod
    def save_results(self,leaderboard: pd.DataFrame, predictions_dict :dict, proba_predictions_dict:dict, feature_importance_dict:dict, y_true:pd.DataFrame, saving_path : Path = None):
        '''Method for saving results returned by the convert() method in a specified or a default directory'''
        pass
    @abstractmethod
    def convert(self, saving_path:Path)->Tuple[pd.DataFrame, dict, dict, dict, pd.DataFrame]:
        '''Primary method performing all the necessary calculations and returning the converted objects.
            1. pd.DataFrame - leaderboard with all models and their scores 
            2. dict : predictions_dict
            3. dict : proba_predictions_dict
            4. dict : feature_importance_dict
            5. pd.DataFrame : y_true 
        '''
        pass







class PredictorConverter(Converter):
    '''
        Class for performing conversion of Autogluon Predictor object to : 

        1. leaderboard dataframe with all tested models and their evaluation scores, where eval_metric is the primarily minimized metric
        2. model - predictions dictionary containing all models and their prediction vectors (as Series)
        3. model -proba_predictions dictionary containing all models and their proba predictions (as pd.DataFrame)
        4. model - list disctionary containing lists of features sorted by their feature importace for each model
    '''

    MULTICLASS_METRICS = list(MULTICLASS_METRICS.keys())
    BINARY_METRICS = list(BINARY_METRICS.keys())
            
    def __init__(self, predictor : TabularPredictor, test_data : TabularDataset, df_name : str, feature_imp_needed = True):
        '''
            Initializes the converter with given TabularPredictor object and a test dataset, 
            all metrics are set based on the problem type of the predictor.

            Args:
                predictor : trained Autogluon predictor object
                test_data : test dataset for evaluation
                df_name : name of the dataset for saving purposes
                feature_imp_needed : boolean value indicating whether the calculation of feature importances is needed
        '''
        if not isinstance(predictor, TabularPredictor):
            raise TypeError(f'predictor argument must be a TabularPredictor object, got {type(predictor).__name__}')
        if not isinstance(test_data, (TabularDataset, pd.DataFrame)):
            raise TypeError(f'test_data argument must be a TabularDataset object, got {type(test_data).__name__}')
        if not isinstance(df_name, str):
            raise TypeError(f'df_name argument must be a string, got {type(df_name).__name__}')
        
        if not predictor.is_fit:
            raise ValueError("The predictor has not been trained. Call predictor.fit() first.")
        if len(test_data)==0:
            raise ValueError('Test dataset is empty. Please provide a valid test data.')
        if not df_name.strip():
            raise ValueError("df_name must be a non-empty string. Please provide a valid df_name.")
        
        self.predictor = predictor
        self.test_data = test_data
        self.df_name = df_name
        self.feature_imp_needed = feature_imp_needed
        
        if predictor.problem_type == 'binary':
            self.metrics = self.BINARY_METRICS
        elif predictor.problem_type == 'multiclass':
            self.metrics = self.MULTICLASS_METRICS
        else:
            raise ValueError("Unsupported problem type. Only 'binary' and 'multiclass' are supported.")
        self.leaderboard = self.create_leaderboard()


    def create_leaderboard(self)-> pd.DataFrame:
        '''
            Method for creating a leaderboard by dropping and renaming columns from a leaderboard returned by the Predictor.

            Returns:
                pd.DataFrame : cleaned leaderboard 
        '''
        leaderboard = self.predictor.leaderboard(self.test_data, extra_metrics=self.metrics)
        leaderboard = leaderboard.drop(columns= ['pred_time_test_marginal','pac','log_loss','mcc', 'quadratic_kappa','pred_time_val_marginal', 'fit_time_marginal', 'stack_level','can_infer', 'fit_order','score_val','score_test','pred_time_test','pred_time_val','fit_time'])
        valid_models = []
        for model_name in leaderboard['model'].tolist():
            try:
                self.predictor.predict(self.test_data.head(1), model=model_name)
                valid_models.append(model_name)
            except AttributeError as e:
                msg = str(e)
                if "'NoneType' object" in msg:
                    logging.warning(f"{YELLOW}Skipping model {model_name}: model is empty or not loaded ({type(e).__name__}: {e}){RESET}")
                else:
                    raise
            except Exception as e:
                raise
        leaderboard = leaderboard[leaderboard['model'].isin(valid_models)].reset_index(drop=True)
        return leaderboard
    
    def create_predictions_dict(self)-> dict:
        '''
            Method for saving predictions of each model in dictionary format

            Returns:
                dict : key - model name, value - predictions for model
        '''
        leaderboard = self.leaderboard

        predictions_dict ={}
        for model in leaderboard['model']:
            predictions = self.predictor.predict(self.test_data, model = model)
            predictions_dict[model] = predictions
        return predictions_dict
    
    def create_proba_predictions_dict(self)-> dict:
        '''
            Method for saving probability predictions of each model in dictionary format

            Returns:
                dict : key - model name, value - probability predictions for model

        '''
        leaderboard = self.leaderboard

        predictions_dict ={}
        for model in leaderboard['model']:
            predictions = self.predictor.predict_proba(self.test_data, model = model)
            predictions_dict[model] = predictions
        return predictions_dict
    
    def create_feature_importance_dict(self) -> dict:
        '''
            Method for obtaining feature importance for each model (where it's possible),
            then sorting features based on their importance and storing in a list format.

            Returns:
                dict: key - model name, value - list of sorted features by feature importance (descending order)
        '''
        leaderboard = self.leaderboard
        importance_dict = {}
        for model in leaderboard['model']:
            feature_importance_df = self.predictor.feature_importance(data = self.test_data, model = model).reset_index()
            features_sorted_list = feature_importance_df.sort_values('importance', ascending=False)['index'].tolist()
            importance_dict[model] = features_sorted_list
        return importance_dict
    
    def extract_target_column(self)-> pd.DataFrame:
        '''
            Method for extracting real y values from the test dataset.
        '''
        target_col = self.predictor.label
        y_test = self.test_data[[target_col]]
        return y_test

    def save_results(self,leaderboard: pd.DataFrame, predictions_dict :dict, proba_predictions_dict:dict, feature_importance_dict:dict, y_true:pd.DataFrame, saving_path : Path = None):
        '''
            Method used for saving .convert() method results on disk.
            If saving_path is None then saving_path by default is converter_results/autogluon/df_name 
        '''
        if saving_path is None:
            saving_path = RES_SAVING_DIR/"autogluon"/self.df_name
        saving_dir = Path(saving_path)
        saving_dir.mkdir(parents=True, exist_ok=True)

        leaderboard.to_csv(saving_dir/'leaderboard.csv', index=False)
        y_true.to_csv(saving_dir/'y_true.csv', index=False)

        with open(saving_dir/'predictions_dict.pkl', 'wb') as f:
            pickle.dump(predictions_dict, f)

        with open(saving_dir/'proba_predictions_dict.pkl', 'wb') as f:
            pickle.dump(proba_predictions_dict, f)

        if self.feature_imp_needed:
            with open(saving_dir/'feature_importance_dict.pkl', 'wb') as f:
                pickle.dump(feature_importance_dict, f)
        print(f'Autogluon conversion results saved to: {saving_dir.resolve()}')


    
    def convert(self, saving_path : Path = None) -> Tuple[pd.DataFrame, Dict, Dict, Dict, pd.DataFrame] :
        '''
            Primary method performing all necessary transformations and saving the results. 

            Args:
                saving_path : if saving_path is None then saving_path by default is converter_results/autogluon/df_name 

            Returns:
                pd.DataFrame : leaderboard created in create_leaderboard() method
                dict : predictions dict created in create_predictions_dict() method
                dict: proba_predictions dict created in create_proba_predictions_dict() method
                dict : feature importance dict created in create_feature_importance_dict() method 
        '''
        leaderboard = self.create_leaderboard()
        predictions_dict = self.create_predictions_dict()
        proba_predictions_dict = self.create_proba_predictions_dict()
        y_true = self.extract_target_column()
        if self.feature_imp_needed:
            feature_importance_dict = self.create_feature_importance_dict()
        else:
            feature_importance_dict = None

        self.save_results(leaderboard, predictions_dict, proba_predictions_dict, feature_importance_dict, y_true, saving_path)
        return leaderboard, predictions_dict, proba_predictions_dict, feature_importance_dict, y_true

        

#-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

class H2O_Converter(Converter):
    '''
        Class for converting H2OAutoML trained models into :
        
        1. leaderboard dataframe with all tested models and their evaluation scores calculated based on the task type (binary or multiclass)
        2. model - predictions dictionary containing all models and their prediction vectors (np.ndarray)
        3. model -proba_predictions dictionary containing all models and their proba predictions (as pd.DataFrame)
        4. model - list disctionary containing lists of features sorted by their feature importace for each model
    '''
    
    MULTICLASS_METRICS_DICT={ 
        'accuracy' : accuracy_score,
        'balanced_accuracy' : balanced_accuracy_score,
        'precision_macro': lambda y_true, y_pred : precision_score(y_true, y_pred, average='macro'),
        'precision_micro' : lambda y_true, y_pred : precision_score(y_true, y_pred, average='micro'),
        'precision_weighted' : lambda y_true, y_pred : precision_score(y_true, y_pred, average='weighted'),
        'recall_macro':lambda y_true, y_pred : recall_score(y_true, y_pred, average='macro'), 
        'recall_micro':lambda y_true, y_pred : recall_score(y_true, y_pred, average='micro'), 
        'recall_weighted':lambda y_true, y_pred : recall_score(y_true, y_pred, average='weighted'),
        'f1_macro': lambda y_true, y_pred : f1_score(y_true, y_pred, average='macro'),
        'f1_micro':lambda y_true, y_pred : f1_score(y_true, y_pred, average='micro'),
        'f1_weighted':lambda y_true, y_pred : f1_score(y_true, y_pred, average='weighted'),
        'roc_auc_ovo':lambda y_true, y_pred, labels : roc_auc_score(y_true, y_pred, multi_class='ovo', labels=labels),
        'roc_auc_ovo_weighted':lambda y_true, y_pred, labels : roc_auc_score(y_true, y_pred, multi_class='ovo', average='weighted',labels=labels),
        'roc_auc_ovr':lambda y_true, y_pred, labels : roc_auc_score(y_true, y_pred, multi_class='ovr',labels=labels),
        'roc_auc_ovr_micro':lambda y_true, y_pred, labels: roc_auc_score(y_true, y_pred, multi_class='ovr', average='micro',labels=labels),
        'roc_auc_ovr_weighted': lambda y_true, y_pred , labels: roc_auc_score(y_true, y_pred, multi_class='ovr', average='weighted',labels=labels)}
    
    BINARY_METRICS_DICT = {
            'accuracy' : accuracy_score,
            'balanced_accuracy' : balanced_accuracy_score,
            'roc_auc' : roc_auc_score,
            'average_precision': lambda y_true,y_pred : average_precision_score(y_true, y_pred, pos_label="1"),
            'precision': lambda y_true,y_pred : precision_score(y_true, y_pred, pos_label="1", zero_division=0), 
            'precision_macro': lambda y_true, y_pred : precision_score(y_true, y_pred, average='macro'),
            'precision_micro' : lambda y_true, y_pred : precision_score(y_true, y_pred, average='micro'),
            'precision_weighted' : lambda y_true, y_pred : precision_score(y_true, y_pred, average='weighted'),
            'recall': lambda y_true,y_pred :recall_score(y_true, y_pred,pos_label="1" ),
            'recall_macro':lambda y_true, y_pred : recall_score(y_true, y_pred, average='macro'), 
            'recall_micro':lambda y_true, y_pred : recall_score(y_true, y_pred, average='micro'), 
            'recall_weighted':lambda y_true, y_pred : recall_score(y_true, y_pred, average='weighted'),
            'f1':  lambda y_true, y_pred : f1_score(y_true, y_pred, pos_label="1"),
            'f1_macro': lambda y_true, y_pred : f1_score(y_true, y_pred, average='macro'),
            'f1_micro':lambda y_true, y_pred : f1_score(y_true, y_pred, average='micro'),
            'f1_weighted':lambda y_true, y_pred : f1_score(y_true, y_pred, average='weighted')}
    
    

    def __init__(self, models_directory : Path, test_data : h2o.H2OFrame, target_column : str, df_name:str, feature_imp_needed = True):
        '''
            Initializes the converter with a given path to a directory containing all trained H2O models and a test dataset, 
            all metrics are set based on the problem type calculated from target column.

            Args:
                models_directory : path to a directory with trained H2O models for a given dataset
                test_data : test dataset for evaluation
                target_column : name of the target column. Used to determine task type.
                df_name : name of the dataset for saving purposes
                feature_imp_needed : boolean value indicating whether the calculation of feature importances is needed
        '''

        if h2o.cluster() is None or not h2o.cluster().is_running():
            h2o.init()

        if not isinstance(models_directory, Path):
            raise TypeError(f'models_directory must be a pathlib.Path object, got {type(models_directory).__name__}')
        if not models_directory.exists():
            raise FileNotFoundError(f'models_directory path does not exist: {models_directory}')
        if not any(models_directory.iterdir()):
            raise ValueError(f'models_directory is empty: {models_directory}')

       
        if not isinstance(test_data, h2o.H2OFrame):
            raise TypeError(f'test_data must be an h2o.H2OFrame object, got {type(test_data).__name__}')
        if test_data.nrow == 0:
            raise ValueError('test_data H2OFrame is empty')

    
        if not isinstance(target_column, str):
            raise TypeError(f'target_column must be a string, got {type(target_column).__name__}')
        if target_column.strip() == "":
            raise ValueError('target_column cannot be empty or whitespace')
        if target_column not in test_data.col_names:
            raise ValueError(f'target_column "{target_column}" not found in test_data columns: {test_data.col_names}')

     
        if not isinstance(df_name, str):
            raise TypeError(f'df_name must be a string, got {type(df_name).__name__}')
        if df_name.strip() == "":
            raise ValueError('df_name cannot be empty or whitespace')

        self.models_directory = models_directory
        self.test_data = test_data
        self.target_column = target_column
        self.df_name = df_name
        self.feature_imp_needed = feature_imp_needed
        self.task_type = self.determine_task_type()

        self.loaded_models = self.load_all_models()
        self.prediction_frames_dict = self.get_prediction_frames()

        self.class_prediction_dict = self.create_predictions_dict()
        self.proba_predictions_dict = self.create_proba_predictions_dict()
        self.classes = sorted(self.test_data[self.target_column].as_data_frame().iloc[:,0].unique().astype(str))

    def determine_task_type(self)->str:
        '''
            Method to establish task type based on the given target column.

            Returns:
                str : task type 'binary' or 'multiclass'
        '''
        target_column_type = self.test_data.types[self.target_column]
        if target_column_type =='enum':
            num_classes = self.test_data[self.target_column].nlevels()[0]
            if num_classes == 2:
                return 'binary'
            elif num_classes>2:
                return 'multiclass'
            else:
                raise ValueError('Less than 2 classes detected in test dataset.')
        else:
            raise ValueError('Task types other than classification are not supported.')
        
        
    def load_all_models(self)->list:
        '''
            Method for loading all models objects saved in models_directory into a list format

            Returns:
                list : List of H2O model objects loaded from the `models_directory`.
        '''
        loaded_models=[]
        for model_file in self.models_directory.iterdir():
            if model_file.is_file():  
                model = h2o.load_model(str(model_file))
                loaded_models.append(model)
        return loaded_models
    
    def get_prediction_frames(self)->dict:
        '''
            Method for accessing every models prediction H2O result and storing results in a model - Frame

            Returns:
                dict : key - model_id , value - predictions Frame
        '''
        predictions = {}
        for model in self.loaded_models:
            predictions_df = model.predict(self.test_data)
            model_id = model.model_id
            predictions[model_id] = predictions_df
        return predictions

    
    def create_predictions_dict(self)->dict:
        '''
            Method for accessing every models prediction vector and storing results in a model - vector dictionary format

            Returns:
                dict : key - model_id , value - predictions vector (ndarray)
        '''
        class_predictions_dict ={}
        for model_id, frame in self.prediction_frames_dict.items():
            predictions_df = frame
            class_predictions = predictions_df['predict'].as_data_frame().iloc[:,0]
            class_predictions_dict[model_id] = class_predictions
        return class_predictions_dict
    
    def create_proba_predictions_dict(self)->Tuple[dict, list]:
        '''
            Method for accessing every models proba predictions  and storing results in a model - prediction matrix dictionary format

            Returns:
                dict : key - model_id , value - predictions matrix (n samples, n classes) pd.DataFrame
        '''
        proba_predictions_dict ={}
        for model_id, frame in self.prediction_frames_dict.items():
            predictions_df = frame
            proba_columns = [col for col in predictions_df.columns if col != 'predict']
            proba_df = predictions_df[proba_columns].as_data_frame()
            if self.task_type =='binary':
                proba_df.columns = [str(i) for i in range(len(proba_columns))] #0-1 column names
                classes = [str(i) for i in range(len(proba_columns))]
            else:
                classes = proba_columns
            
            proba_predictions_dict[model_id] = proba_df
          
        return proba_predictions_dict
    
    def create_feature_importance_dict(self)->dict:
        '''
            Method for obtaining feature importance (for models where it's possible) and sorting features based on their importance.

            Returns:
                dict : key - model_id, value - sorted list of features or None for models where feature importance can't be obtained
        '''
        importance_dict ={}
        for model in self.loaded_models:
            try:
                importance_df = model.varimp(use_pandas=True)
                sorted_variables = importance_df.sort_values('scaled_importance', ascending=False)['variable'].tolist()
                importance_dict[model.model_id] = sorted_variables
            except (AttributeError, NotImplementedError):
                print(f'Model {model.model_id} does not have .varimp(). None inserted.')
                importance_dict[model.model_id] = None
        return importance_dict
    
    def extract_target_column(self)->pd.DataFrame:
        test_data_df = self.test_data.as_data_frame()
        y_true = test_data_df[[self.target_column]]
        return y_true

    
    def calculate_multiclass_metrics(self) -> pd.DataFrame:
        '''
            Method for calculating all evaluation metrics for multiclass task type for each model. 

            Returns:
                pd.DataFrame : DataFrame with models as rows and columns as their evaluation metrics
        '''
        
        y_true_raw = self.test_data[self.target_column].as_data_frame().iloc[:,0].to_numpy()

        encoder = LabelEncoder()
        encoder.fit(self.classes)
        y_true_encoded = encoder.transform(y_true_raw)

        rows=[]
        for model_id in self.class_prediction_dict.keys():
            row = {'model': model_id}
            y_pred_class = self.class_prediction_dict[model_id]
            y_pred_encoded_class = encoder.transform(y_pred_class)
            y_pred_proba = self.proba_predictions_dict[model_id].to_numpy()

            for metric_name, metric_func in self.MULTICLASS_METRICS_DICT.items():
                try:
                    #roc auc base on probabilities 
                    if metric_name.startswith('roc_auc'):
                        row[metric_name] = metric_func(y_true_encoded, y_pred_proba, labels = range(len(self.classes)))
                    else: #other metrics based on class predictions (encoded by LabelEncoder just in case)
                        row[metric_name] = metric_func(y_true_encoded, y_pred_encoded_class)
                except Exception as e:
                    row[metric_name] = None
                    print(f'Error with {metric_name} : {e}')
            rows.append(row)
        metrics_df = pd.DataFrame(rows)
        return metrics_df
    
    def calculate_binary_metrics(self)->pd.DataFrame:
        '''
            Method for calculating all evaluation metrics for binary classification task type for each model. 

            Returns:
                pd.DataFrame : DataFrame with models as rows and columns as their evaluation metrics
        '''
        y_true_raw = self.test_data[self.target_column].as_data_frame().iloc[:,0]
        y_true = y_true_raw.astype(str).to_numpy() # classes '0' and '1'

        rows=[]
        for model_id in self.class_prediction_dict.keys():
            row = {'model': model_id}

            y_class_pred = self.class_prediction_dict[model_id].astype(str) 
            y_pred_proba = self.proba_predictions_dict[model_id].to_numpy()

            for metric_name, metric_func in self.BINARY_METRICS_DICT.items():
                try:
                    
                    #roc auc and avg precision need positive class probability:
                    if metric_name =='roc_auc' or metric_name=='average_precision':
                        row[metric_name] = metric_func(y_true, y_pred_proba[:,1])
                    else:
                        row[metric_name] = metric_func(y_true, y_class_pred)

                except Exception as e:
                    row[metric_name] = None
                    print(f'Error with metric: {metric_name} : {e}')
            rows.append(row)
        metrics_df = pd.DataFrame(rows)
        return metrics_df
    
    def save_results(self,leaderboard: pd.DataFrame, predictions_dict :dict, proba_predictions_dict:dict, feature_importance_dict:dict, y_true: pd.DataFrame, saving_path : Path = None):
        '''
            Method used for saving .convert() results on disk.
            If saving_path is None then saving_path by default is converter_results/h2o/df_name 
        '''
        if saving_path is None:
            saving_path  = RES_SAVING_DIR/"h2o"/self.df_name
        saving_dir = Path(saving_path)
        saving_dir.mkdir(parents=True, exist_ok=True)

        leaderboard.to_csv(saving_dir/'leaderboard.csv', index=False)
        y_true.to_csv(saving_dir/'y_true.csv', index=False)

        with open(saving_dir/'predictions_dict.pkl', 'wb') as f:
            pickle.dump(predictions_dict, f)

        with open(saving_dir/'proba_predictions_dict.pkl', 'wb') as f:
            pickle.dump(proba_predictions_dict, f)
        
        if self.feature_imp_needed:
            with open(saving_dir/'feature_importance_dict.pkl', 'wb') as f:
                pickle.dump(feature_importance_dict, f)
        print(f'H2O conversion results saved to: {saving_dir.resolve()}')
    

    def convert(self, saving_path : Path = None) -> Tuple[pd.DataFrame, dict, dict, dict, pd.DataFrame]:
        '''
            Primary method used to perform all calculations. 

            Args:
                saving_path : if saving_path is None then saving_path by default is converter_results/h2o/df_name 

            Returns:
                pd.DataFrame : leaderboard obtained by calculating evaluation metrics for the given task type 
                dict : predictions dict returned by get_class_predictions_dict() method
                dict : proba_predictions dict returned by get_proba_predictions_dict() method
                dict : feature importance dict returned by get_feature_importance_dict() method
        '''
        if self.task_type == 'binary':
            leaderboard = self.calculate_binary_metrics()
        else:
            leaderboard = self.calculate_multiclass_metrics()
        if self.feature_imp_needed:
            importance_dict = self.create_feature_importance_dict()
        else:
            importance_dict = None
        y_true = self.extract_target_column()

        self.save_results(leaderboard, self.class_prediction_dict, self.proba_predictions_dict, importance_dict,y_true, saving_path)
        
        return leaderboard, self.class_prediction_dict, self.proba_predictions_dict, importance_dict, y_true



from datetime import datetime
import pandas as pd
import h2o
from h2o.frame import H2OFrame
import random
import inspect 
import os
import sys
import importlib.resources as pkg_resources
import subprocess
import time
import psutil
from pathlib import Path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
from arsa_ml.rashomon_set import *
from arsa_ml.rashomon_intersection import *
from arsa_ml.visualizers.rashomon_visualizer import *
from arsa_ml.visualizers.intersection_visualizer import *
from app.rashomon_binary_app import *
from arsa_ml.converters import *
from arsa_ml.pipelines.builder_abstract import *
import pickle
import warnings
warnings.filterwarnings("ignore")
import logging
BLUE_BOLD = "\033[1;34m"
RESET = "\033[0m"

logging.basicConfig(
    level=logging.INFO,
    format=f"{BLUE_BOLD}%(message)s{RESET}"
)

PROJECT_ROOT = Path(__file__).resolve().parents[3] #pipelines -> rahsomon_analysis ->src -> root
# RASHOMON_BINARY_APP_PATH = PROJECT_ROOT/"src"/"app"/"rashomon_binary_app.py"
# RASHOMON_MULTICLASS_APP_PATH = PROJECT_ROOT/"src"/"app"/"rashomon_multiclass_app.py"
# INTERSECTION_BINARY_APP_PATH = PROJECT_ROOT/"src"/"app"/"intersection_binary_app.py"
# INTERSECTION_MULTICLASS_APP_PATH = PROJECT_ROOT/"src"/"app"/"intersection_multiclass_app.py"

INTERSECTION_BINARY_APP_PATH = Path(pkg_resources.files("app") / "intersection_binary_app.py")
INTERSECTION_MULTICLASS_APP_PATH = Path(pkg_resources.files("app") / "intersection_multiclass_app.py")
RASHOMON_BINARY_APP_PATH = Path(pkg_resources.files("app") / "rashomon_binary_app.py")
RASHOMON_MULTICLASS_APP_PATH = Path(pkg_resources.files("app") / "rashomon_multiclass_app.py")


PIPELINES_DIR = Path(__file__).resolve().parent
TMP_DIR = PIPELINES_DIR / "tmp"
TMP_DIR.mkdir(exist_ok=True)  



class BuildRashomonH2O(Builder):
    '''
        Pipeline for creating and exploring the Rashomon Set from user-provided models from H2O framework. 
    '''

    def __init__(self, models_directory: Path, test_data : h2o.H2OFrame, target_column : str, df_name : str, 
                 base_metric : str, delta : float = 0.1, feature_imp_needed = True,  converter_results_directory : Path = None):
        '''
            Initializes the Pipeline user's input (models traines with H2O framework) to the creation of the Rashomon Set and it's visualizations.

            Args:
                models_directory : path to a folder where all models saved from H2O output are stored.
                test_data : test dataset for evaluation
                target_column : name of the target column. Used to determine task type.
                df_name : name of the dataset for saving converter output purposes.
                base_metric : evaluation metric to be used while creating the Rashomon Set
                delta : delta parameter for probabilistic ambiguity and discrepancy (used only for binary task type). If not specified the default value of 0.1 will be used.
                feature_imp_needed : boolean value indicating whether the calculation of feature importances is needed.
                converter_results_directory : Path to directory where converter outputs will be stored.
                                              If None -> creates 'converter_results/<df_name>_<timestamp>' next to models_directory.
        '''

        #All __init__ parameter validation is done in other modules, so it’s not necessary to duplicate it here.
        logging.info("Converting H2O models to Rashomon Analysis format...")
        if converter_results_directory is None:
            timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            converter_results_directory = models_directory.parent/'converter_results'/f"{df_name}_{timestamp}"
        logging.info(f"Converter results will be stored in: {converter_results_directory}")
        logging.info("Please wait...")
        logging.info("This process may take several minutes depending on the number of models and dataset size.")
        logging.info("If you wish to speed up the process, consider setting feature_imp_needed=False.")
        h2o_converter = H2O_Converter(models_directory, test_data, target_column, df_name, feature_imp_needed)
        self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.y_true = h2o_converter.convert(saving_path = converter_results_directory)
        self.base_metric = base_metric
        self.delta = delta
        self.epsilon = None
        logging.info("Conversion completed.")
        logging.info("You can preview the Rashomon Set size plot using preview_rashomon() method to determine epsilon parameter value.")

    
    def preview_rashomon(self):
        '''
            Method illustrating the leaderboard and the plot with all possible epsilon values and the Rashomon Set sizes to help choose the right epsilon argument.
        '''

        print("\033[1;34mThe Visualizer received a leaderboard displayed below.\033[0m")
        print("\033[1;34mCheck if this is the desired leaderboard for analysis, else change the given Path parameter\033[0m")
        print(self.leaderboard)
        print("\033[1;34mThe plot illustrating the Rashomon Set size for different epsilon values will be displayed....\033[0m")

        self.visualize_rashomon_set_volume()


    def visualize_rashomon_set_volume(self):
        '''
            Method for visualising Rashomon set size depending on epsilon.
        '''

        epsilons = np.linspace(0, 1, 100)
        tmp_rashomon_set = RashomonSet(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.base_metric, 10.0)
        valid_epsilons = [eps for eps in epsilons if len(tmp_rashomon_set.get_rashomon_set(eps))>1]

        def compute_size(epsilon):
            rs = RashomonSet(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.base_metric, epsilon)
            return len(rs.rashomon_set)
            
        rashomon_sizes= [compute_size(eps) for eps in valid_epsilons]
        # jump epsilons
        diff = np.diff(rashomon_sizes)
        jump_indices = np.where(diff > 0)[0] + 1

        plt.figure(figsize=(8,5))
        plt.scatter(valid_epsilons, rashomon_sizes, s=20, color = '#6d9c5d')
        plt.gca().yaxis.set_major_locator(MaxNLocator(integer=True))

        for idx in jump_indices:
            jump_epsilon = valid_epsilons[idx]
            plt.axvline(x=jump_epsilon, color='#eea951', linestyle='--', alpha=0.8)
            plt.text(jump_epsilon, plt.ylim()[0] - 0.2, f"{jump_epsilon:.3f}",
             rotation=90, color='#eea951', fontsize=9, ha='center', va='top')

        if len(jump_indices) > 0:
            plt.axvline(x=valid_epsilons[jump_indices[0]], color='#eea951', linestyle='--', alpha=0.5,
                        label='Change in Rashomon set size')
        plt.legend(fontsize=10)
        plt.xlabel('Epsilon value', labelpad=19)
        plt.ylabel('Number of models in Rashomon set')
        plt.title('Rashomon set sizes', pad=10)
        plt.tight_layout()
        plt.show()


    def set_epsilon(self, epsilon: float):
        '''
            Sets the epsilon threshold for the Rashomon Set.
            Epsilon value must be set before calling build() method.
        '''

        if epsilon>=0:
            self.epsilon = epsilon
        else:
            raise ValueError("Epsilon parameter cannot be less than 0. Please provide a valid epsilon value.")


    def build(self, launch_dashboard: bool = True):
        '''
            Builds the Rashomon Set pipeline from H2O models.
            Creates Rashomon Set object and Visualizer object from user's input and launches a Streamlit dashboard in a subprocess for interactive visualization.

            Parameters: 
                launch_dashboard : boolean value indicating whether to launch the Streamlit dashboard after building the Rashomon Set and Visualizer. Default is True.
            
            Returns:
                    RashomonSet : Rashomon Set class object computed during pipeline
                    RashomonVisualizer : Visualizer class object created for computed Rashomon Set
        '''

        logging.info("Building Rashomon Set and Visualizer...")
        logging.info('Validating epsilon parameter...')
        if self.epsilon is None:
            raise ValueError("Please set the epsilon value using the set_epsilon() method. For help call rashomon_preview() method.")
        logging.info(f"Epsilon parameter set to {self.epsilon}.")

        rashomon_set = RashomonSet(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.base_metric, self.epsilon)
        visualizer = Visualizer(rashomon_set, self.y_true)

        if launch_dashboard:
            plots ={}
            descriptions ={}

            if rashomon_set.task_type == "binary":
                method_names = visualizer.binary_methods
                ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr = visualizer.lolipop_ambiguity_discrepancy_proba_version(self.delta)
                plots["lolipop_ambiguity_discrepancy_proba_version"], descriptions["lolipop_ambiguity_discrepancy_proba_version"] = ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr
                proba_ambiguity_plot, proba_ambiguity_descr = visualizer.proba_ambiguity_vs_epsilon(self.delta)
                plots["proba_ambiguity_vs_epsilon"], descriptions["proba_ambiguity_vs_epsilon"] = proba_ambiguity_plot, proba_ambiguity_descr
                proba_discrepancy_plot, proba_discrepancy_descr = visualizer.proba_discrepancy_vs_epsilon(self.delta)
                plots["proba_discrepancy_vs_epsilon"], descriptions["proba_discrepancy_vs_epsilon"] = proba_discrepancy_plot, proba_discrepancy_descr

            elif rashomon_set.task_type =="multiclass":
                method_names = visualizer.multiclass_methods

            random_idx = random.choice(self.y_true.index.tolist()) #choose random sample for analysis
            for method in method_names:
                func = getattr(visualizer, method)
                sig = inspect.signature(func)
                params = sig.parameters

                #if the method needs parameters (random sample index):
                if len(params)>0:
                    # random index 
                    if "sample_index" in params:
                        plot, descr = func(sample_index=random_idx)
                    else: 
                        raise ValueError(f"Method {method} needs unsupported parameters")
                else: plot, descr = func()
                plots[method] = plot
                descriptions[method] = descr
            
            logging.info("Rashomon Set and Visualizer built successfully.")
            temp_file = TMP_DIR / "temp_plots.pkl"
            with open(temp_file, "wb") as f:
                pickle.dump((plots, descriptions), f)

            logging.info('Closing all previous Streamlit processes...')
            self.dashboard_close() # close all processes before starting a new one
            logging.info('Launching Streamlit dashboard...')
            if rashomon_set.task_type =="binary":           
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(RASHOMON_BINARY_APP_PATH), "--", str(temp_file)
                ])
            elif rashomon_set.task_type=="multiclass":
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(RASHOMON_MULTICLASS_APP_PATH), "--", str(temp_file)
                ])

            logging.info(f"Streamlit dashboard launched (PID={proc.pid})")
            logging.info("You can continue working with your code - to close Streamlit process call dashboard_close() method.")
        else: 
            logging.info("Rashomon Set and Visualizer built successfully. Dashboard launch skipped as per user request.")

        return rashomon_set, visualizer
    
    def dashboard_close(self):
        '''
            Method for stopping all Streamlit processes and closing the dashboard.
        '''

        for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
            try:
                cmd = " ".join(proc.info['cmdline'] or [])
                if 'streamlit' in cmd:
                    print(f"🛑 Killing Streamlit (PID={proc.info['pid']})")
                    psutil.Process(proc.info['pid']).terminate()
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue
    

class BuildRashomonIntersectionH2O(Builder):
    '''
        Pipeline for creating and exploring the Rashomon Set Intersection from user-provided models from H2O framework. 
    '''

    def __init__(self, models_directory: Path, test_data : h2o.H2OFrame, target_column : str, df_name : str, 
                 metrics : list, custom_weights : list = None, weighted_sum_method : str = 'entropy', delta : float = 0.1, 
                 feature_imp_needed = True,  converter_results_directory : Path = None):
        '''
            Initializes the Pipeline user's input (models traines with H2O framework) to the creation of the Rashomon Set Intersection and it's visualizations.

            Args:
                models_directory : path to a folder where all models saved from H2O output are stored.
                test_data : test dataset for evaluation
                target_column : name of the target column. Used to determine task type.
                df_name : name of the dataset for saving converter output purposes.
                metrics : list of metrics to be used in the intersection calculation
                custom_weights : if weighted_sum is 'custom_weights' then user must specify weights in 2-element list.
                weighted_sum_method : Specifies the method or weights for selecting the base model. Options are:
                    - None or 'entropy' (default): Uses entropy-based method to select the base model.
                    - 'critic': Uses critic-based method to select the base model.
                    - 'custom_weights': Selects the base model using a weights provided by user.
                    the base model as a weighted combination of the two metrics.
                delta : delta parameter for probabilistic ambiguity and discrepancy (used only for binary task type). If not specified the default value of 0.1 will be used.
                feature_imp_needed : boolean value indicating whether the calculation of feature importances is needed.
                converter_results_directory : Path to directory where converter outputs will be stored.
                                              If None -> creates 'converter_results/<df_name>_<timestamp>' next to models_directory.
            
        '''

        #All __init__ parameter validation is done in other modules, so it’s not necessary to duplicate it here.
        logging.info("Converting H2O models to Rashomon Intersection Analysis format...")
        if converter_results_directory is None:
            timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            converter_results_directory = models_directory.parent/'converter_results'/f"{df_name}_{timestamp}"
        logging.info(f"Converter results will be stored in: {converter_results_directory}")
        logging.info("Please wait...")
        logging.info("This process may take several minutes depending on the number of models and dataset size.")
        logging.info("If you wish to speed up the process, consider setting feature_imp_needed=False.")
        h2o_converter = H2O_Converter(models_directory, test_data, target_column, df_name, feature_imp_needed)
        self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.y_true = h2o_converter.convert(saving_path = converter_results_directory)
        self.metrics = metrics
        self.custom_weights = custom_weights
        self.weighted_sum_method = weighted_sum_method
        self.delta = delta
        self.epsilon = None
        logging.info("Conversion completed.")
        logging.info("You can preview the Rashomon Intersection Set size plot using preview_rashomon() method to determine epsilon parameter value.")

    
    def preview_rashomon(self):
        '''
            Method illustrating the leaderboard and the plot with all possible epsilon values and the Rashomon Intersection sizes to help choose the right epsilon argument.
        '''

        print("\033[1;34mThe Visualizer received a leaderboard displayed below.\033[0m")
        print("\033[1;34mCheck if this is the desired leaderboard for analysis, else change the given Path parameter\033[0m")
        print(self.leaderboard)
        print("\033[1;34mThe plot illustrating the Rashomon Set size for different epsilon values will be displayed....\033[0m")

        self.visualize_rashomon_set_volume()


    def visualize_rashomon_set_volume(self):
        '''
            Method for visualising Rashomon Set size depending on epsilon
        '''

        epsilons = np.linspace(0, 1, 100)
        tmp_rashomon_int = RashomonIntersection(leaderboard=self.leaderboard, predictions=self.predictions, proba_predictions=self.proba_predictions, 
                                               feature_importances=self.feature_importances, metrics=self.metrics, epsilon=10.0, 
                                               custom_weights=self.custom_weights, weighted_sum_method=self.weighted_sum_method)
        valid_epsilons = [eps for eps in epsilons if len(tmp_rashomon_int.get_rashomon_set(eps))>1]

        def compute_size(epsilon):
            rs = RashomonIntersection(leaderboard=self.leaderboard, predictions=self.predictions, proba_predictions=self.proba_predictions, 
                                               feature_importances=self.feature_importances, metrics=self.metrics, epsilon=epsilon, 
                                               custom_weights=self.custom_weights, weighted_sum_method=self.weighted_sum_method)
            return len(rs.rashomon_set)
            
        rashomon_sizes= [compute_size(eps) for eps in valid_epsilons]
        # jump epsilons
        diff = np.diff(rashomon_sizes)
        jump_indices = np.where(diff > 0)[0] + 1

        plt.figure(figsize=(8,5))
        plt.scatter(valid_epsilons, rashomon_sizes, s=20, color = '#6d9c5d')
        plt.gca().yaxis.set_major_locator(MaxNLocator(integer=True))

        for idx in jump_indices:
            jump_epsilon = valid_epsilons[idx]
            plt.axvline(x=jump_epsilon, color='#eea951', linestyle='--', alpha=0.8)
            plt.text(jump_epsilon, plt.ylim()[0] - 0.2, f"{jump_epsilon:.3f}",
             rotation=90, color='#eea951', fontsize=9, ha='center', va='top')

        if len(jump_indices) > 0:
            plt.axvline(x=valid_epsilons[jump_indices[0]], color='#eea951', linestyle='--', alpha=0.5,
                        label='Change in Rashomon set size')
        plt.legend(fontsize=10)
        plt.xlabel('Epsilon value', labelpad=19)
        plt.ylabel('Number of models in Rashomon set')
        plt.title('Rashomon set sizes', pad=10)
        plt.tight_layout()
        plt.show()


    def set_epsilon(self, epsilon: float):
        '''
            Sets the epsilon threshold for the Rashomon Intersection.
            Epsilon value must be set before calling build() method.
        '''

        if epsilon>=0:
            self.epsilon = epsilon
        else:
            raise ValueError("Epsilon parameter cannot be less than 0. Please provide a valid epsilon value.")


    def build(self, launch_dashboard: bool = True):
        '''
            Builds the Rashomon Intersection pipeline from H2O models.
            Creates Rashomon Intersection object and Intersection Visualizer object from user's input and launches a Streamlit dashboard in a subprocess for interactive visualization.

            Parameters:
                launch_dashboard : boolean value indicating whether to launch the Streamlit dashboard after building the Rashomon Intersection and Visualizer. Default is True.
                
            Returns:
                    RashomonSetIntersection : Rashomon Set Intersection class object computed during pipeline
                    IntersectionVisualizer : Visualizer class object created for computed Rashomon Set Intersection
        '''

        logging.info("Building Rashomon Intersection and Visualizer...")
        logging.info('Validating epsilon parameter...')
        if self.epsilon is None:
            raise ValueError("Please set the epsilon value using the set_epsilon() method. For help call rashomon_preview() method.")
        logging.info(f"Epsilon parameter set to {self.epsilon}.")
        
        rashomon_set = RashomonIntersection(leaderboard=self.leaderboard, predictions=self.predictions, proba_predictions=self.proba_predictions, 
                                               feature_importances=self.feature_importances, metrics=self.metrics, epsilon=self.epsilon, 
                                               custom_weights=self.custom_weights, weighted_sum_method=self.weighted_sum_method)
        visualizer = IntersectionVisualizer(rashomon_set, self.y_true)
        
        if launch_dashboard:
            plots ={}
            descriptions ={}

            if rashomon_set.task_type == "binary":
                method_names = visualizer.binary_methods
                ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr = visualizer.lolipop_ambiguity_discrepancy_proba_version(self.delta)
                plots["lolipop_ambiguity_discrepancy_proba_version"], descriptions["lolipop_ambiguity_discrepancy_proba_version"] = ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr
                proba_ambiguity_plot, proba_ambiguity_descr = visualizer.proba_ambiguity_vs_epsilon(self.delta)
                plots["proba_ambiguity_vs_epsilon"], descriptions["proba_ambiguity_vs_epsilon"] = proba_ambiguity_plot, proba_ambiguity_descr
                proba_discrepancy_plot, proba_discrepancy_descr = visualizer.proba_discrepancy_vs_epsilon(self.delta)
                plots["proba_discrepancy_vs_epsilon"], descriptions["proba_discrepancy_vs_epsilon"] = proba_discrepancy_plot, proba_discrepancy_descr

            elif rashomon_set.task_type =="multiclass":
                method_names = visualizer.multiclass_methods

            random_idx = random.choice(self.y_true.index.tolist()) #choose random sample for analysis
            for method in method_names:
                func = getattr(visualizer, method)
                sig = inspect.signature(func)
                params = sig.parameters

                #if the method needs parameters (random sample index):
                if len(params)>0:
                    # random index 
                    if "sample_index" in params:
                        plot, descr = func(sample_index=random_idx)
                    else: 
                        raise ValueError(f"Method {method} needs unsupported parameters")
                else: plot, descr = func()
                plots[method] = plot
                descriptions[method] = descr
            
            logging.info("Rashomon Intersection Set and Visualizer built successfully.")
            temp_file = TMP_DIR / "temp_plots.pkl"
            with open(temp_file, "wb") as f:
                pickle.dump((plots, descriptions), f)

            logging.info('Closing all previous Streamlit processes...')
            self.dashboard_close() # close all processes before starting a new one
            logging.info('Launching Streamlit dashboard...')

            if rashomon_set.task_type =="binary":           
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(INTERSECTION_BINARY_APP_PATH), "--", str(temp_file)
                ])
            elif rashomon_set.task_type=="multiclass":
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(INTERSECTION_MULTICLASS_APP_PATH), "--", str(temp_file)
                ])

            logging.info(f"Streamlit dashboard launched (PID={proc.pid})")
            logging.info("You can continue working with your code - to close Streamlit process call dashboard_close() method.")
        else:
            logging.info("Rashomon Intersection and Visualizer built successfully. Dashboard launch skipped as per user request.")

        return rashomon_set, visualizer
    
    def dashboard_close(self):
        '''
            Method for stopping all Streamlit processes and closing the dashboard.
        '''
        for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
            try:
                cmd = " ".join(proc.info['cmdline'] or [])
                if 'streamlit' in cmd:
                    print(f"🛑 Killing Streamlit (PID={proc.info['pid']})")
                    psutil.Process(proc.info['pid']).terminate()
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue
    

class BuildRashomonAutogluon(Builder):
    '''
        Pipeline for creating and exploring the Rashomon Set from user-provided models from AutoGluon framework. 
    '''
    
    def __init__(self, predictor : TabularPredictor, test_data : TabularDataset, df_name : str, base_metric : str,
                delta : float = 0.1, feature_imp_needed = True,  converter_results_directory : Path = None):
        '''
            Initializes the Pipeline user's input (models traines with AutoGluon framework) to the creation of the Rashomon Set and it's visualizations.

            Args:
                predictor : trained Autogluon predictor object
                test_data : test dataset for evaluation
                df_name : name of the dataset for saving converter output purposes.
                base_metric : evaluation metric to be used while creating the Rashomon Set
                delta : delta parameter for probabilistic ambiguity and discrepancy (used only for binary task type). If not specified the default value of 0.1 will be used.
                feature_imp_needed : boolean value indicating whether the calculation of feature importances is needed.
                converter_results_directory : Path to directory where converter outputs will be stored.
                                              If None -> creates 'converter_results/<df_name>_<timestamp>' in the current directory.
        '''

        #All __init__ parameter validation is done in other modules, so it’s not necessary to duplicate it here.
        logging.info("Converting AutoGluon models to Rashomon Analysis format...")
        if converter_results_directory is None:
            timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            converter_results_directory = Path.cwd()/'converter_results'/f"{df_name}_{timestamp}"
        logging.info(f"Converter results will be stored in: {converter_results_directory}")
        logging.info("Please wait...")
        logging.info("This process may take several minutes depending on the number of models and dataset size.")
        autogluon_converter = PredictorConverter(predictor=predictor, test_data=test_data, df_name=df_name, feature_imp_needed=feature_imp_needed)
        self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.y_true = autogluon_converter.convert(saving_path = converter_results_directory)
        self.base_metric = base_metric
        self.delta = delta
        self.epsilon = None
        logging.info("Conversion completed.")
        logging.info("You can preview the Rashomon Set size plot using preview_rashomon() method to determine epsilon parameter value.")


    
    def preview_rashomon(self):
        '''
            Method illustrating the leaderboard and the plot with all possible epsilon values and the Rashomon Set sizes to help choose the right epsilon argument.
        '''

        print("\033[1;34mThe Visualizer received a leaderboard displayed below.\033[0m")
        print("\033[1;34mCheck if this is the desired leaderboard for analysis, else change the given Path parameter\033[0m")
        print(self.leaderboard)
        print("\033[1;34mThe plot illustrating the Rashomon Set size for different epsilon values will be displayed....\033[0m")

        self.visualize_rashomon_set_volume()


    def visualize_rashomon_set_volume(self):
        '''
            Method for visualising Rashomon set size depending on epsilon
        '''

        epsilons = np.linspace(0, 1, 100)
        tmp_rashomon_set = RashomonSet(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.base_metric, 10.0)
        valid_epsilons = [eps for eps in epsilons if len(tmp_rashomon_set.get_rashomon_set(eps))>1]

        def compute_size(epsilon):
            rs = RashomonSet(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.base_metric, epsilon)
            return len(rs.rashomon_set)
            
        rashomon_sizes= [compute_size(eps) for eps in valid_epsilons]
        # jump epsilons
        diff = np.diff(rashomon_sizes)
        jump_indices = np.where(diff > 0)[0] + 1

        plt.figure(figsize=(8,5))
        plt.scatter(valid_epsilons, rashomon_sizes, s=20, color = '#6d9c5d')
        plt.gca().yaxis.set_major_locator(MaxNLocator(integer=True))

        for idx in jump_indices:
            jump_epsilon = valid_epsilons[idx]
            plt.axvline(x=jump_epsilon, color='#eea951', linestyle='--', alpha=0.8)
            plt.text(jump_epsilon, plt.ylim()[0] - 0.2, f"{jump_epsilon:.3f}",
             rotation=90, color='#eea951', fontsize=9, ha='center', va='top')

        if len(jump_indices) > 0:
            plt.axvline(x=valid_epsilons[jump_indices[0]], color='#eea951', linestyle='--', alpha=0.5,
                        label='Change in Rashomon set size')
        plt.legend(fontsize=10)
        plt.xlabel('Epsilon value', labelpad=19)
        plt.ylabel('Number of models in Rashomon set')
        plt.title('Rashomon set sizes', pad=10)
        plt.tight_layout()
        plt.show()


    def set_epsilon(self, epsilon: float):
        '''
            Sets the epsilon threshold for the Rashomon Set.
            Epsilon value must be set before calling build() method.
        '''

        if epsilon>=0:
            self.epsilon = epsilon
        else:
            raise ValueError("Epsilon parameter cannot be less than 0. Please provide a valid epsilon value.")


    def build(self, launch_dashboard: bool = True):
        '''
            Builds the Rashomon Set pipeline from AutoGluon input.
            Creates Rashomon Set object and Visualizer object from user's input and launches a Streamlit dashboard in a subprocess for interactive visualization.

            Parameters:
                launch_dashboard : boolean value indicating whether to launch the Streamlit dashboard after building the Rashomon Set and Visualizer. Default is True.
                
            Returns:
                    RashomonSet : Rashomon Set class object computed during pipeline
                    RashomonVisualizer : Visualizer class object created for computed Rashomon Set
        '''

        logging.info("Building Rashomon Set and Visualizer...")
        logging.info('Validating epsilon parameter...')
        if self.epsilon is None:
            raise ValueError("Please set the epsilon value using the set_epsilon() method. For help call rashomon_preview() method.")
        logging.info(f"Epsilon parameter set to {self.epsilon}.")
        
        rashomon_set = RashomonSet(self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.base_metric, self.epsilon)
        visualizer = Visualizer(rashomon_set, self.y_true)
        
        if launch_dashboard:
            plots ={}
            descriptions ={}

            if rashomon_set.task_type == "binary":
                method_names = visualizer.binary_methods
                ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr = visualizer.lolipop_ambiguity_discrepancy_proba_version(self.delta)
                plots["lolipop_ambiguity_discrepancy_proba_version"], descriptions["lolipop_ambiguity_discrepancy_proba_version"] = ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr
                proba_ambiguity_plot, proba_ambiguity_descr = visualizer.proba_ambiguity_vs_epsilon(self.delta)
                plots["proba_ambiguity_vs_epsilon"], descriptions["proba_ambiguity_vs_epsilon"] = proba_ambiguity_plot, proba_ambiguity_descr
                proba_discrepancy_plot, proba_discrepancy_descr = visualizer.proba_discrepancy_vs_epsilon(self.delta)
                plots["proba_discrepancy_vs_epsilon"], descriptions["proba_discrepancy_vs_epsilon"] = proba_discrepancy_plot, proba_discrepancy_descr

            elif rashomon_set.task_type =="multiclass":
                method_names = visualizer.multiclass_methods

            random_idx = random.choice(self.y_true.index.tolist()) #choose random sample for analysis
            for method in method_names:
                    func = getattr(visualizer, method)
                    sig = inspect.signature(func)
                    params = sig.parameters

                    #if the method needs parameters (random sample index):
                    if len(params)>0:
                        # random index 
                        if "sample_index" in params:
                            plot, descr = func(sample_index=random_idx)
                        else: 
                            raise ValueError(f"Method {method} needs unsupported parameters")
                    else: plot, descr = func()
                    plots[method] = plot
                    descriptions[method] = descr
            
            logging.info("Rashomon Set and Visualizer built successfully.")
            temp_file = TMP_DIR / "temp_plots.pkl"
            with open(temp_file, "wb") as f:
                pickle.dump((plots, descriptions), f)

            logging.info('Closing all previous Streamlit processes...')
            self.dashboard_close() # close all processes before starting a new one
            logging.info('Launching Streamlit dashboard...')

            if rashomon_set.task_type =="binary":           
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(RASHOMON_BINARY_APP_PATH), "--", str(temp_file)
                ])
            elif rashomon_set.task_type=="multiclass":
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(RASHOMON_MULTICLASS_APP_PATH), "--", str(temp_file)
                ])

            logging.info(f"Streamlit dashboard launched (PID={proc.pid})")
            logging.info("You can continue working with your code - to close Streamlit process call dashboard_close() method.")
        
        else:
            logging.info("Rashomon Set and Visualizer built successfully. Dashboard launch skipped as per user request.")

        return rashomon_set, visualizer

    def dashboard_close(self):
        '''
            Method for stopping all Streamlit processes and closing the dashboard.
        '''
        for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
            try:
                cmd = " ".join(proc.info['cmdline'] or [])
                if 'streamlit' in cmd:
                    print(f"🛑 Killing Streamlit (PID={proc.info['pid']})")
                    psutil.Process(proc.info['pid']).terminate()
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue
    
class BuildRashomonIntersectionAutogluon(Builder):
    '''
        Pipeline for creating and exploring the Rashomon Set Intersection from user-provided models from Autogluon framework. 
    '''

    def __init__(self, predictor : TabularPredictor, test_data : TabularDataset, df_name : str, metrics : list, 
                 custom_weights : list = None, weighted_sum_method : str = 'entropy', delta : float = 0.1, 
                 feature_imp_needed = True,  converter_results_directory : Path = None):
        '''
            Initializes the Pipeline user's input (models traines with AutoGluon framework) to the creation of the Rashomon Set Intersection and it's visualizations.

            Args:
                predictor : trained Autogluon predictor object
                test_data : test dataset for evaluation
                df_name : name of the dataset for saving converter output purposes.
                metrics : list of metrics to be used in the intersection calculation
                custom_weights : if weighted_sum is 'custom_weights' then user must specify weights in 2-element list.
                weighted_sum_method : Specifies the method or weights for selecting the base model. Options are:
                    - None or 'entropy' (default): Uses entropy-based method to select the base model.
                    - 'critic': Uses critic-based method to select the base model.
                    - 'custom_weights': Selects the base model using a weights provided by user.
                    the base model as a weighted combination of the two metrics.
                delta : delta parameter for probabilistic ambiguity and discrepancy (used only for binary task type). If not specified the default value of 0.1 will be used.
                feature_imp_needed : boolean value indicating whether the calculation of feature importances is needed.
                converter_results_directory : Path to directory where converter outputs will be stored.
                                              If None -> creates 'converter_results/<df_name>_<timestamp>' in the current directory.
            
        '''

        #All __init__ parameter validation is done in other modules, so it’s not necessary to duplicate it here.
        logging.info("Converting AutoGluon models to Rashomon Intersection Analysis format...")
        if converter_results_directory is None:
            timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            converter_results_directory = Path.cwd()/'converter_results'/f"{df_name}_{timestamp}"
        logging.info(f"Converter results will be stored in: {converter_results_directory}")
        logging.info("Please wait...")
        logging.info("This process may take several minutes depending on the number of models and dataset size.")
        logging.info("If you wish to speed up the process, consider setting feature_imp_needed=False.") 
        autogluon_converter = PredictorConverter(predictor=predictor, test_data=test_data, df_name=df_name, feature_imp_needed=feature_imp_needed)
        self.leaderboard, self.predictions, self.proba_predictions, self.feature_importances, self.y_true = autogluon_converter.convert(saving_path = converter_results_directory)
        self.metrics = metrics
        self.custom_weights = custom_weights
        self.weighted_sum_method = weighted_sum_method
        self.delta = delta
        self.epsilon = None
        logging.info("Conversion completed.")
        logging.info("You can preview the Rashomon Intersection Set size plot using preview_rashomon() method to determine epsilon parameter value.")

    
    def preview_rashomon(self):
        '''
            Method illustrating the leaderboard and the plot with all possible epsilon values and the Rashomon Intersection sizes to help choose the right epsilon argument.
        '''

        print("\033[1;34mThe Visualizer received a leaderboard displayed below.\033[0m")
        print("\033[1;34mCheck if this is the desired leaderboard for analysis, else change the given Path parameter\033[0m")
        print(self.leaderboard)
        print("\033[1;34mThe plot illustrating the Rashomon Set size for different epsilon values will be displayed....\033[0m")

        self.visualize_rashomon_set_volume()


    def visualize_rashomon_set_volume(self):
        '''
            Method for visualising Rashomon set size depending on epsilon
        '''

        epsilons = np.linspace(0, 1, 100)
        tmp_rashomon_int = RashomonIntersection(leaderboard=self.leaderboard, predictions=self.predictions, proba_predictions=self.proba_predictions, 
                                               feature_importances=self.feature_importances, metrics=self.metrics, epsilon=10.0, 
                                               custom_weights=self.custom_weights, weighted_sum_method=self.weighted_sum_method)
        valid_epsilons = [eps for eps in epsilons if len(tmp_rashomon_int.get_rashomon_set(eps))>1]

        def compute_size(epsilon):
            rs = RashomonIntersection(leaderboard=self.leaderboard, predictions=self.predictions, proba_predictions=self.proba_predictions, 
                                               feature_importances=self.feature_importances, metrics=self.metrics, epsilon=epsilon, 
                                               custom_weights=self.custom_weights, weighted_sum_method=self.weighted_sum_method)
            return len(rs.rashomon_set)
            
        rashomon_sizes= [compute_size(eps) for eps in valid_epsilons]
        # jump epsilons
        diff = np.diff(rashomon_sizes)
        jump_indices = np.where(diff > 0)[0] + 1

        plt.figure(figsize=(8,5))
        plt.scatter(valid_epsilons, rashomon_sizes, s=20, color = '#6d9c5d')
        plt.gca().yaxis.set_major_locator(MaxNLocator(integer=True))

        for idx in jump_indices:
            jump_epsilon = valid_epsilons[idx]
            plt.axvline(x=jump_epsilon, color='#eea951', linestyle='--', alpha=0.8)
            plt.text(jump_epsilon, plt.ylim()[0] - 0.2, f"{jump_epsilon:.3f}",
             rotation=90, color='#eea951', fontsize=9, ha='center', va='top')

        if len(jump_indices) > 0:
            plt.axvline(x=valid_epsilons[jump_indices[0]], color='#eea951', linestyle='--', alpha=0.5,
                        label='Change in Rashomon set size')
        plt.legend(fontsize=10)
        plt.xlabel('Epsilon value', labelpad=19)
        plt.ylabel('Number of models in Rashomon set')
        plt.title('Rashomon set sizes', pad=10)
        plt.tight_layout()
        plt.show()


    def set_epsilon(self, epsilon: float):
        '''
            Sets the epsilon threshold for the Rashomon Intersection.
            Epsilon value must be set before calling build() method.
        '''

        if epsilon>=0:
            self.epsilon = epsilon
        else:
            raise ValueError("Epsilon parameter cannot be less than 0. Please provide a valid epsilon value.")


    def build(self, launch_dashboard : bool = True):
        '''
            Builds the Rashomon Intersection pipeline from AutoGluon input.
            Creates Rashomon Intersection object and Intersection Visualizer object from user's input and launches a Streamlit dashboard in a subprocess for interactive visualization.

            Parameters:
                launch_dashboard : boolean value indicating whether to launch the Streamlit dashboard after building the Rashomon
                    
            Returns:
                    RashomonSetIntersection : Rashomon Set Intersection class object computed during pipeline
                    IntersectionVisualizer : Visualizer class object created for computed Rashomon Set Intersection
        '''

        logging.info("Building Rashomon Intersection and Visualizer...")
        logging.info('Validating epsilon parameter...')
        if self.epsilon is None:
            raise ValueError("Please set the epsilon value using the set_epsilon() method. For help call rashomon_preview() method.")
        logging.info(f"Epsilon parameter set to {self.epsilon}.")
        
        rashomon_set = RashomonIntersection(leaderboard=self.leaderboard, predictions=self.predictions, proba_predictions=self.proba_predictions, 
                                               feature_importances=self.feature_importances, metrics=self.metrics, epsilon=self.epsilon, 
                                               custom_weights=self.custom_weights, weighted_sum_method=self.weighted_sum_method)
        visualizer = IntersectionVisualizer(rashomon_set, self.y_true)

        if launch_dashboard:
            plots ={}
            descriptions ={}

            if rashomon_set.task_type == "binary":
                method_names = visualizer.binary_methods
                ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr = visualizer.lolipop_ambiguity_discrepancy_proba_version(self.delta)
                plots["lolipop_ambiguity_discrepancy_proba_version"], descriptions["lolipop_ambiguity_discrepancy_proba_version"] = ambiguity_discrepancy_proba_plot, ambiguity_discrepancy_proba_descr
                proba_ambiguity_plot, proba_ambiguity_descr = visualizer.proba_ambiguity_vs_epsilon(self.delta)
                plots["proba_ambiguity_vs_epsilon"], descriptions["proba_ambiguity_vs_epsilon"] = proba_ambiguity_plot, proba_ambiguity_descr
                proba_discrepancy_plot, proba_discrepancy_descr = visualizer.proba_discrepancy_vs_epsilon(self.delta)
                plots["proba_discrepancy_vs_epsilon"], descriptions["proba_discrepancy_vs_epsilon"] = proba_discrepancy_plot, proba_discrepancy_descr

            elif rashomon_set.task_type =="multiclass":
                method_names = visualizer.multiclass_methods

            random_idx = random.choice(self.y_true.index.tolist()) #choose random sample for analysis
            for method in method_names:
                    func = getattr(visualizer, method)
                    sig = inspect.signature(func)
                    params = sig.parameters

                    #if the method needs parameters (random sample index):
                    if len(params)>0:
                        # random index 
                        if "sample_index" in params:
                            plot, descr = func(sample_index=random_idx)
                        else: 
                            raise ValueError(f"Method {method} needs unsupported parameters")
                    else: plot, descr = func()
                    plots[method] = plot
                    descriptions[method] = descr
            
            logging.info("Rashomon Intersection Set and Visualizer built successfully.")
            temp_file = TMP_DIR / "temp_plots.pkl"
            with open(temp_file, "wb") as f:
                pickle.dump((plots, descriptions), f)

            logging.info('Closing all previous Streamlit processes...')
            self.dashboard_close()
            logging.info('Launching Streamlit dashboard...')

            if rashomon_set.task_type =="binary":           
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(INTERSECTION_BINARY_APP_PATH), "--", str(temp_file)
                ])
            elif rashomon_set.task_type=="multiclass":
                proc = subprocess.Popen([
                    sys.executable, "-m", "streamlit", "run", str(INTERSECTION_MULTICLASS_APP_PATH), "--", str(temp_file)
                ])

            logging.info(f"Streamlit dashboard launched (PID={proc.pid})")
            logging.info("You can continue working with your code - to close Streamlit process call dashboard_close() method.")
        else:
            logging.info("Rashomon Set and Visualizer built successfully. Dashboard launch skipped as per user request.")

        return rashomon_set, visualizer

    def dashboard_close(self):
        '''
            Method for stopping all Streamlit processes and closing the dashboard.
        '''
        for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
            try:
                cmd = " ".join(proc.info['cmdline'] or [])
                if 'streamlit' in cmd:
                    print(f"🛑 Killing Streamlit (PID={proc.info['pid']})")
                    psutil.Process(proc.info['pid']).terminate()
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue




import streamlit as st
import sys
import pickle
import os
import plotly.io as pio
from pathlib import Path
import streamlit.components.v1 as components
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from arsa_ml.visualizers.rashomon_visualizer import *
from arsa_ml import *
BASE_DIR = Path(__file__).resolve().parent.parent
assets_dir = BASE_DIR / "style"

def load_converterd_results(df_name, framework_name):
    leaderboard = pd.read_csv( BASE_DIR.parent / "converter_results" / framework_name/ df_name / "leaderboard.csv")
    y_true = pd.read_csv( BASE_DIR.parent / "converter_results" / framework_name/ df_name / "y_true.csv")
    with open(BASE_DIR.parent / "converter_results" / framework_name / df_name / "predictions_dict.pkl", "rb") as f:
        predictions_dict = pickle.load(f)

    with open(BASE_DIR.parent / "converter_results" / framework_name / df_name / "proba_predictions_dict.pkl", "rb") as f:
        proba_predictions_dict = pickle.load(f)

    with open(BASE_DIR.parent / "converter_results" / framework_name / df_name / "feature_importance_dict.pkl", "rb") as f:
        feature_importance_dict = pickle.load(f)

    return leaderboard, predictions_dict, proba_predictions_dict, feature_importance_dict, y_true


def main():
    print(str(assets_dir))
    l, p_d, proba_d, f_d, y = load_converterd_results('compas', 'h2o')
    rashomon = RashomonSet(l, p_d, proba_d, f_d, 'accuracy', 0.1)
    rashomon.summarize_rashomon()
    

    # visualizer = Visualizer(rashomon, y)
    # fig, descr = visualizer.vpr_vs_base_model_plot()
    # st.plotly_chart(fig, use_container_width=True)
    # st.markdown(descr)

    # ratios, ratios_descr = visualizer.pattern_ratio_indicator()
    # st.plotly_chart(ratios)
    # st.markdown(ratios_descr, unsafe_allow_html=True)

    # ratios, ratios_descr = visualizer.rashomon_ratio_indicator()
    # st.plotly_chart(ratios)
    # st.markdown(ratios_descr, unsafe_allow_html=True)

    # table_rashomon, table_descr = visualizer.generate_rashomon_set_table()
    # st.dataframe(table_rashomon)
    # st.markdown(table_descr, unsafe_allow_html=True)

    #capacity_by_class, capacity_by_class_descr = visualizer.rashomon_capacity_distribution_by_class()
    #st.plotly_chart(capacity_by_class)
    #st.markdown(capacity_by_class_descr, unsafe_allow_html=True)

    #capacity_by_class, capacity_by_class_descr = visualizer.rashomon_capacity_distribution()
    #t.plotly_chart(capacity_by_class)
    #st.markdown(capacity_by_class_descr, unsafe_allow_html=True)

    #capacity_by_class, capacity_by_class_descr = visualizer.rashomon_capacity_distribution_by_class()
    #st.plotly_chart(capacity_by_class)
    #t.markdown(capacity_by_class_descr, unsafe_allow_html=True)

    #table_rashomon, table_descr = visualizer.feature_importance_table()
    #st.plotly_chart(table_rashomon)
    #st.markdown(table_descr, unsafe_allow_html=True)

    #fi_heatmap, fi_heatmap_descr = visualizer.feature_importance_heatmap()
    #st.plotly_chart(fi_heatmap)
    #st.markdown(fi_heatmap_descr, unsafe_allow_html=True)
    
    # test działania wykresów w połączniu z html css
   
    # '''
    # html_fig = pio.to_html(fig, full_html=False, config={'responsive': True})

    # with open(assets_dir / "layout.html") as f:
    #     layout_html = f.read()

    # layout_html = layout_html.replace("{{FIGURE}}", html_fig)
    # components.html(layout_html, height = 800, scrolling=False)
    # '''


if __name__ == "__main__":
    main()


import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
from arsa_ml.rashomon_set import *
from arsa_ml.rashomon_intersection import *
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import pandas as pd
from arsa_ml.visualizers.rashomon_visualizer import *

plot_title_font = dict(
    color="#6B7F8A",
    family="Inter, Segoe UI, sans-serif",  
    size=16,
    style="normal",
    weight=400            
)

color_palette = ['#fcfafa', '#6C929E','#94B4C1','#205781', '#132440']
chart_colors_v2 = [
    '#DBE8EF',  
    '#C8DCEA',  
    '#B5CDD9',  
    '#9FBDCF',  
    '#89ADC4',  
    '#739DB9',  
    '#638EA8',  
    '#557D95',  
    '#476C82',  
    '#3A5B70',
    "#16476A",
    "#6C929E"
]

contrast_colors = [ "#ed801a"]
class_colors = ['#205781', '#739DB9', '#94B4C1', '#9FBDCF']

class_colors = [
    '#08306b',  
    '#08519c',  
    '#2171b5',
    '#4292c6',
    '#6baed6',
    '#9ecae1',
    '#9fbdcf',
    '#739db9',
    '#94b4c1',
    '#deebf7'  
]


class IntersectionVisualizer(Visualizer):
    '''
        Child class of the Visualizer. Used to generate plots for the Rashomon Intersection analysis.

        Arguments:
            rashomon_intersection: RashomonIntersection object to be analyzed
            y_true : true prediction vector

        Attributes:
            rashomon_set : passed RashomonIntersection object
            y_true : passed y_true object
            binary_methods : plots that can be generated for the binary classification task type
            multiclass_methods: plots that can be generated fot the multiclass classification task type
    '''

    def __init__(self, rashomon_intersection : RashomonIntersection, y_true : pd.DataFrame):
        self.rashomon_set = rashomon_intersection
        self.y_true = y_true

        self.binary_methods = ["generate_rashomon_set_table","base_model_return","weighted_method_return","base_metric_return", "number_of_classes_return", "set_size_indicator","pareto_front_plot","generate_gauge_for_metrics","plot_venn_diagram","rashomon_ratio_indicator", "pattern_ratio_indicator","lolipop_ambiguity_discrepancy","ambiguity_vs_epsilon", "discrepancy_vs_epsilon",
                                            "rashomon_ratio_vs_epsilon","pattern_rashomon_ratio_vs_epsilon", "rashomon_capacity_distribution","rashomon_capacity_distribution_by_class",
                                            "percent_agreement_barplot","vprs_widths_plot", "vpr_width_histogram", "feature_importance_table", "feature_importance_heatmap",
                                            "agreement_rates_density", "vpr_vs_base_model_plot","proba_probabilities_for_sample", "rashomon_capacity_for_sample", "cohens_kappa_heatmap", "cohens_kappa_diverging_barplot"] 
        self.multiclass_methods =  ["generate_rashomon_set_table","base_model_return","weighted_method_return","base_metric_return", "number_of_classes_return","set_size_indicator","pareto_front_plot", "generate_gauge_for_metrics","plot_venn_diagram","rashomon_ratio_indicator", "pattern_ratio_indicator","lolipop_ambiguity_discrepancy","ambiguity_vs_epsilon", "discrepancy_vs_epsilon",
                                            "rashomon_ratio_vs_epsilon","pattern_rashomon_ratio_vs_epsilon", "rashomon_capacity_distribution","rashomon_capacity_distribution_by_class",
                                            "percent_agreement_barplot", "feature_importance_table", "feature_importance_heatmap",
                                            "agreement_rates_density","proba_probabilities_for_sample", "rashomon_capacity_for_sample", "cohens_kappa_heatmap", "cohens_kappa_diverging_barplot"]
        
    #override 
    def base_metric_return(self)->Tuple[go.Figure, str]:
        '''Overrides the base_metric_return method from Visualiser by returning both evaluation metrics from the RashomonIntersection object'''
        return go.Figure(), self.rashomon_set.metrics
    
    def weighted_method_return(self)->Tuple[go.Figure, str]:
        '''Returns the name of the selected weighted_sum_method from the RashomonIntersection and an empty plot'''
        return go.Figure(), self.rashomon_set.weighted_sum_method
    
    #override
    def set_size_indicator(self)->Tuple[go.Figure, str]:
        '''Overrides the set_size_indicator method from Visualiser by returning the gauge plot with the size of the RashomonIntersection'''
        rashomon_size = len(self.rashomon_set.rashomon_set)
        all_models_num = len(self.rashomon_set.leaderboard)
        fig = go.Figure(go.Indicator(
            mode = "number + gauge",
            value = rashomon_size,
            number = {'font': {'size': 40, 'color': '#6B7F8A'}},
            title =dict(text = 'Number of models in the Rashomon Intersection',
                        font = plot_title_font),
            gauge = {
                'axis': {'range': [0, all_models_num], 'dtick': 5, 'tickfont': {'color': '#6B7F8A'}},
                'bar': {'color': chart_colors_v2[6]},
                'steps': [
                    {'range': [0, all_models_num], 'color': "#ebebeb"},
                ],
            }
        ))

        return fig, "The gauge plot represents how many models from the leaderboard were included in the Rashomon Intersection for the given parameters."
    

    #override
    def ambiguity_vs_epsilon(self)-> Tuple[go.Figure, str]:
        '''
            Method for creating a plot representing possible ambiguity values for different epsilon paramaters and it's description.
            
            Returns:
                Tuple[go.Figure, str]
                - plot
                - plot description
        '''
        epsilons = np.linspace(0, 1, 100)
        valid_epsilons = [eps for eps in epsilons if len(self.rashomon_set.get_rashomon_set(eps))>1]

        def compute_ambiguity(epsilon):
            rs = RashomonIntersection(self.rashomon_set.leaderboard, self.rashomon_set.predictions_dict, self.rashomon_set.proba_predictions_dict, self.rashomon_set.feature_importance_dict, self.rashomon_set.metrics, epsilon, self.rashomon_set.custom_weights, self.rashomon_set.weighted_sum_method)
            if rs.task_type=='binary':
                return rs.binary_ambiguity()
            elif rs.task_type=='multiclass':
                return rs.multiclass_ambiguity()
            else:
                raise ValueError("Unknown task type for the Rashomon Set")
            
        ambiguity_vals = [compute_ambiguity(eps) for eps in valid_epsilons]

        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x = valid_epsilons, 
            y = ambiguity_vals, 
            mode='markers+lines',
            marker = dict(color=chart_colors_v2[11], size=2),
            line=dict(color=chart_colors_v2[11], width=2),
            text=[f'Epsilon: {x:.2f}<br>Ambiguity: {y:.2f}' for x, y in zip(valid_epsilons, ambiguity_vals)],
            hoverinfo='text',
            showlegend=False
        ))

        if self.rashomon_set.task_type=='binary':
            real_ambiguity = self.rashomon_set.binary_ambiguity()
        elif self.rashomon_set.task_type=='multiclass':
            real_ambiguity = self.rashomon_set.multiclass_ambiguity()

        fig.add_trace(go.Scatter(
            x=[self.rashomon_set.epsilon],
            y=[real_ambiguity],
            mode='markers',
            marker=dict(color=contrast_colors[0], size=5),
            text = [f'Epsilon: {self.rashomon_set.epsilon:.2f} <br> Ambiguity : {real_ambiguity:.2f}'],
            hoverinfo="text",
            name = "Ambiguity for the given Rashomon Set"
        ))
        step = max(1, len(valid_epsilons)//10)
        tickvals = [round(e, 2) for e in valid_epsilons[::step]]

        tickfont_combined = {**plot_axis_font, "size": 9}
        fig.update_layout(
            title = dict(text = 'Ambiguity vs Epsilon values',
                        font = plot_title_font,
                        x=0.5, 
                        xanchor='center'),
            height=400, 
            width = 600,
            xaxis = dict(title=dict(
                            text="Epsilon",
                            font=plot_axis_font),
                         showgrid=False, 
                         tickmode='array', 
                         tickvals = tickvals, 
                         tickangle=45, 
                         tickfont=tickfont_combined),
            yaxis=dict(title=dict(
                            text="Ambiguity",
                            font=plot_axis_font),
                       range = [0,1.05], 
                       showgrid=False, 
                       tickfont=tickfont_combined),
            margin= dict(l=70, r = 40, t = 60, b = 60),
            legend=dict(
                font=dict(size=9),
                x=1,             
                y=0,             
                xanchor='right',
                yanchor='bottom',
                orientation='v',
                bgcolor="rgba(0,0,0,0)"
            ))
        
        descr= """
            The plot shows how Ambiguity changes with epsilon values. The highlighted point is the Ambiguity calculated for the given Rashomon Set. 
        """

        return fig, descr
    

    #override
    def discrepancy_vs_epsilon(self)-> Tuple[go.Figure, str]:
        '''
            Method for creating a plot representing possible discrepancy values for different epsilon paramaters. 
            
            Returns:
                Tuple[go.Figure, str]
                - plot
                - plot description
        '''

        epsilons = np.linspace(0, 1, 100)
        valid_epsilons = [eps for eps in epsilons if len(self.rashomon_set.get_rashomon_set(eps))>1]

        def compute_discrepancy(epsilon):
            rs =RashomonIntersection(self.rashomon_set.leaderboard, self.rashomon_set.predictions_dict, self.rashomon_set.proba_predictions_dict, self.rashomon_set.feature_importance_dict, self.rashomon_set.metrics, epsilon, self.rashomon_set.custom_weights, self.rashomon_set.weighted_sum_method)
            if rs.task_type=='binary':
                return rs.binary_discrepancy()
            elif rs.task_type=='multiclass':
                return rs.multiclass_discrepancy()
            else:
                raise ValueError("Unknown task type for the Rashomon Set")
            
        discrepancy_vals = [compute_discrepancy(eps) for eps in valid_epsilons]

        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x = valid_epsilons, 
            y = discrepancy_vals, 
            mode='markers+lines',
            marker = dict(color=chart_colors_v2[11], size=2),
            line=dict(color=chart_colors_v2[11], width=2),
            text=[f'Epsilon: {x:.2f}<br>Discrepancy: {y:.2f}' for x, y in zip(valid_epsilons, discrepancy_vals)],
            hoverinfo='text',
            showlegend=False
        ))

        if self.rashomon_set.task_type=='binary':
            real_dicrepancy = self.rashomon_set.binary_discrepancy()
        elif self.rashomon_set.task_type=='multiclass':
            real_dicrepancy = self.rashomon_set.multiclass_discrepancy()

        fig.add_trace(go.Scatter(
            x=[self.rashomon_set.epsilon],
            y=[real_dicrepancy],
            mode='markers',
            marker=dict(color=contrast_colors[0], size=5),
            text = [f'Epsilon: {self.rashomon_set.epsilon:.2f} <br> Discrepancy : {real_dicrepancy:.2f}'],
            hoverinfo="text",
            name = "Discrepancy for the given Rashomon Set"
        ))
        step = max(1, len(valid_epsilons)//10)
        tickvals = [round(e, 2) for e in valid_epsilons[::step]]

        tickfont_combined = {**plot_axis_font, "size": 9}
        fig.update_layout(
            title = dict(text = 'Discrepancy vs Epsilon values',
                        font = plot_title_font,
                        x=0.5, 
                        xanchor='center'),
            height=400, 
            width = 600,
            xaxis = dict(title=dict(
                            text="Epsilon",
                            font=plot_axis_font),
                         showgrid=False, 
                         tickmode='array', 
                         tickvals = tickvals, 
                         tickangle=45, 
                         tickfont=tickfont_combined),
            yaxis=dict(title=dict(
                            text="Discrepancy",
                            font=plot_axis_font),
                       range = [0,1.05], 
                       showgrid=False, 
                       tickfont=tickfont_combined),
            margin= dict(l=70, r = 40, t = 60, b = 60),
            #legend in the bottom right corner
            legend=dict(
                font=dict(size=9),
                x=1,             
                y=0,             
                xanchor='right',
                yanchor='bottom',
                orientation='v',
                bgcolor="rgba(0,0,0,0)"
            ))
        
        descr= """
            The plot shows how Discrepancy changes with epsilon values. The highlighted point is the Discrepancy calculated for the given Rashomon Set. 
        """

        return fig, descr
    
    #override
    def proba_ambiguity_vs_epsilon(self, delta : float)-> Tuple[go.Figure, str]:
        '''
            Method for creating a plot representing possible Probabilistic Ambiguity values for different epsilon paramaters and it's description.
            
            Returns:
                Tuple[go.Figure, str]
                - plot
                - plot description
        '''
        epsilons = np.linspace(0, 1, 100)
        valid_epsilons = [eps for eps in epsilons if len(self.rashomon_set.get_rashomon_set(eps))>1]

        def compute_proba_ambiguity(epsilon, delta):
            rs =RashomonIntersection(self.rashomon_set.leaderboard, self.rashomon_set.predictions_dict, self.rashomon_set.proba_predictions_dict, self.rashomon_set.feature_importance_dict, self.rashomon_set.metrics, epsilon, self.rashomon_set.custom_weights, self.rashomon_set.weighted_sum_method)
            if rs.task_type=='binary':
                return rs.probabilistic_abiguity(delta)
            else:
                raise ValueError("Probabilistic Ambiguity cannot be applied to multiclass classification task.")
            
        ambiguity_vals = [compute_proba_ambiguity(eps, delta) for eps in valid_epsilons]

        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x = valid_epsilons, 
            y = ambiguity_vals, 
            mode='markers+lines',
            marker = dict(color=chart_colors_v2[11], size=2),
            line=dict(color=chart_colors_v2[11], width=2),
            text=[f'Epsilon: {x:.2f}<br>Probabilistic Ambiguity: {y:.2f}' for x, y in zip(valid_epsilons, ambiguity_vals)],
            hoverinfo='text',
            showlegend=False
        ))

        if self.rashomon_set.task_type=='binary':
            real_ambiguity = self.rashomon_set.probabilistic_abiguity(delta)
        else:
            raise ValueError("Probabilistic Ambiguity cannot be applied to multiclass classification task.")
            

        fig.add_trace(go.Scatter(
            x=[self.rashomon_set.epsilon],
            y=[real_ambiguity],
            mode='markers',
            marker=dict(color=contrast_colors[0], size=5),
            text = [f'Epsilon: {self.rashomon_set.epsilon:.2f} <br> Probabilistic Ambiguity : {real_ambiguity:.2f}'],
            hoverinfo="text",
            name = "Probabilistic Ambiguity for the given Rashomon Set"
        ))
        step = max(1, len(valid_epsilons)//10)  
        tickvals = [round(e, 2) for e in valid_epsilons[::step]]

        tickfont_combined = {**plot_axis_font, "size": 9}
        fig.update_layout(
            title = dict(text = f'Probabilistic Ambiguity (delta = {delta}) vs Epsilon values',
                        font = plot_title_font,
                        x=0.5, 
                        xanchor='center'),
            height=400, 
            width = 600,
            xaxis = dict(title=dict(
                            text="Epsilon",
                            font=plot_axis_font),
                         showgrid=False, 
                         tickmode='array', 
                         tickvals = tickvals, 
                         tickangle=45, 
                         tickfont=tickfont_combined),
            yaxis=dict(title=dict(
                            text="Probabilistic Ambiguity",
                            font=plot_axis_font),
                       range = [0,1.05], 
                       showgrid=False, 
                       tickfont=tickfont_combined),
            margin= dict(l=70, r = 40, t = 60, b = 60),
            legend=dict(
                font=dict(size=9),
                x=1,             
                y=0,             
                xanchor='right',
                yanchor='bottom',
                orientation='v',
                bgcolor="rgba(0,0,0,0)"
            ))
        
        descr= """
            The plot shows how Probabilistic Ambiguity changes with epsilon values. The highlighted point is the Probabilistic Ambiguity calculated for the given Rashomon Set. 
        """

        return fig, descr
    
    #override
    def proba_discrepancy_vs_epsilon(self, delta : float)-> Tuple[go.Figure, str]:
        '''
            Method for creating a plot representing possible Probabilistic Discrepancy values for different epsilon paramaters. 
            
            Returns:
                Tuple[go.Figure, str]
                - plot
                - plot description
        '''
        epsilons = np.linspace(0, 1, 100)
        valid_epsilons = [eps for eps in epsilons if len(self.rashomon_set.get_rashomon_set(eps))>1]

        def compute_proba_discrepancy(epsilon, delta):
            rs =RashomonIntersection(self.rashomon_set.leaderboard, self.rashomon_set.predictions_dict, self.rashomon_set.proba_predictions_dict, self.rashomon_set.feature_importance_dict, self.rashomon_set.metrics, epsilon, self.rashomon_set.custom_weights, self.rashomon_set.weighted_sum_method)
            if rs.task_type=='binary':
                return rs.probabilistic_discrepancy(delta)
            else:
                raise ValueError("Probabilistic Discrepancy cannot be applied to multiclass classification task.")
            
        discrepancy_vals = [compute_proba_discrepancy(eps, delta) for eps in valid_epsilons]

        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x = valid_epsilons, 
            y = discrepancy_vals, 
            mode='markers+lines',
            marker = dict(color=chart_colors_v2[11], size=2),
            line=dict(color=chart_colors_v2[11], width=2),
            text=[f'Epsilon: {x:.2f}<br>Probabilistic Discrepancy: {y:.2f}' for x, y in zip(valid_epsilons, discrepancy_vals)],
            hoverinfo='text',
            showlegend=False
        ))

        if self.rashomon_set.task_type=='binary':
            real_dicrepancy = self.rashomon_set.probabilistic_discrepancy(delta)
        else:
            raise ValueError("Probabilistic Discrepancy cannot be applied to multiclass classification task.")

        fig.add_trace(go.Scatter(
            x=[self.rashomon_set.epsilon],
            y=[real_dicrepancy],
            mode='markers',
            marker=dict(color=contrast_colors[0], size=5),
            text = [f'Epsilon: {self.rashomon_set.epsilon:.2f} <br>Probabilistic Discrepancy : {real_dicrepancy:.2f}'],
            hoverinfo="text",
            name = "Probabilistic Discrepancy for the given Rashomon Set"
        ))
        step = max(1, len(valid_epsilons)//10)  
        tickvals = [round(e, 2) for e in valid_epsilons[::step]]

        tickfont_combined = {**plot_axis_font, "size": 9}
        fig.update_layout(
            title = dict(text = f'Probabilistic Discrepancy (delta = {delta}) vs Epsilon values',
                        font = plot_title_font,
                        x=0.5, 
                        xanchor='center'),
            height=400, 
            width = 600,
            xaxis = dict(title=dict(
                            text="Epsilon",
                            font=plot_axis_font),
                         showgrid=False, 
                         tickmode='array', 
                         tickvals = tickvals, 
                         tickangle=45, 
                         tickfont=tickfont_combined),
            yaxis=dict(title=dict(
                            text="Probabilistic Discrepancy",
                            font=plot_axis_font),
                       range = [0,1.05], 
                       showgrid=False, 
                       tickfont=tickfont_combined),
            margin= dict(l=70, r = 40, t = 60, b = 60),
            #legend in the bottom right corner
            legend=dict(
                font=dict(size=9),
                x=1,             
                y=0,             
                xanchor='right',
                yanchor='bottom',
                orientation='v',
                bgcolor="rgba(0,0,0,0)"
            ))
        
        descr= """
            The plot shows how Probabilistic Discrepancy changes with epsilon values. The highlighted point is the Probabilistic Discrepancy calculated for the given Rashomon Set. 
        """
        return fig, descr
    
    #override
    def rashomon_ratio_vs_epsilon(self) -> Tuple[go.Figure, str]:
        '''
            Method for creating a plot representing possible Rashomon Ratio values for different epsilon paramaters. 
            
            Returns:
                Tuple[go.Figure, str]
                - plot
                - plot description
        '''
        epsilons = np.linspace(0, 1, 100)
        valid_epsilon = [eps for eps in epsilons if len(self.rashomon_set.get_rashomon_set(eps))>1]

        def compute_rashomon_ratio(epsilon):
            rs = RashomonIntersection(self.rashomon_set.leaderboard, self.rashomon_set.predictions_dict, self.rashomon_set.proba_predictions_dict, self.rashomon_set.feature_importance_dict, self.rashomon_set.metrics, epsilon, self.rashomon_set.custom_weights, self.rashomon_set.weighted_sum_method)
            rashomon_ratio = rs.rashomon_ratio()
            return rashomon_ratio
        
        rashomon_ratio_vals = [compute_rashomon_ratio(epsilon=eps) for eps in valid_epsilon]
        real_rashomon_ratio = self.rashomon_set.rashomon_ratio()

        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x = valid_epsilon, 
            y = rashomon_ratio_vals, 
            mode='markers+lines',
            marker = dict(color=chart_colors_v2[11], size=2),
            line=dict(color=chart_colors_v2[11], width=2),
            text=[f'Epsilon: {x:.2f}<br>Rashomon Ratio: {y:.2f}' for x, y in zip(valid_epsilon, rashomon_ratio_vals)],
            hoverinfo='text',
            showlegend=False
        ))

        fig.add_trace(go.Scatter(
            x=[self.rashomon_set.epsilon],
            y=[real_rashomon_ratio],
            mode='markers',
            marker=dict(color=contrast_colors[0], size=5),
            text = [f'Epsilon: {self.rashomon_set.epsilon:.2f} <br> Rashomon Ratio : {real_rashomon_ratio:.2f}'],
            hoverinfo="text",
            name = "Rashomon Ratio for the given Rashomon Set."
        ))

        step = max(1, len(valid_epsilon)//10)
        tickvals = [round(e, 2) for e in valid_epsilon[::step]]

        tickfont_combined = {**plot_axis_font, "size": 9}
        fig.update_layout(
            title = dict(text = 'Rashomon Ratio vs Epsilon values',
                        font = plot_title_font,
                        x=0.5, 
                        xanchor='center'),
            height=400, 
            width = 600,
            xaxis = dict(title=dict(
                            text="Epsilon",
                            font=plot_axis_font),
                         showgrid=False, 
                         tickmode='array', 
                         tickvals = tickvals, 
                         tickangle=45, 
                         tickfont=tickfont_combined),
            yaxis=dict(title=dict(
                            text="Rashomon Ratio",
                            font=plot_axis_font),
                       range = [0,1.05], 
                       showgrid=False, 
                       tickfont=tickfont_combined),
            margin= dict(l=70, r = 40, t = 60, b = 60),
            #legend in the bottom right corner
            legend=dict(
                font=dict(size=9),
                x=1,             
                y=0,             
                xanchor='right',
                yanchor='bottom',
                orientation='v'
            ))


        descr= """
                The plot shows how Rashomon Ratio changes with epsilon values. The highlighted point is the Rashomon Ratio calculated for the given Rashomon Set. 
            """
        
        return fig, descr
    

    #override
    def pattern_rashomon_ratio_vs_epsilon(self) -> Tuple[go.Figure, str]:
        '''
            Method for creating a plot representing possible Pattern Rashomon Ratio values for different epsilon paramaters. 
            Returns:
                go.Figure plot 
                plot description : str
        '''
        epsilons = np.linspace(0, 1, 100)
        valid_epsilon = [eps for eps in epsilons if len(self.rashomon_set.get_rashomon_set(eps))>1]

        def compute_pattern_rashomon_ratio(epsilon):
            rs = RashomonIntersection(self.rashomon_set.leaderboard, self.rashomon_set.predictions_dict, self.rashomon_set.proba_predictions_dict, self.rashomon_set.feature_importance_dict, self.rashomon_set.metrics, epsilon, self.rashomon_set.custom_weights, self.rashomon_set.weighted_sum_method)
            return rs.pattern_rashomon_ratio()
        
        pattern_rashomon_ratio_vals = [compute_pattern_rashomon_ratio(epsilon=eps) for eps in valid_epsilon]
        real_pattern_rashomon_ratio = self.rashomon_set.pattern_rashomon_ratio()

        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x = valid_epsilon, 
            y = pattern_rashomon_ratio_vals, 
            mode='markers+lines',
            marker = dict(color=chart_colors_v2[11], size=2),
            line=dict(color=chart_colors_v2[11], width=2),
            text=[f'Epsilon: {x:.2f}<br>Pattern Rashomon Ratio: {y:.2f}' for x, y in zip(valid_epsilon, pattern_rashomon_ratio_vals)],
            hoverinfo='text',
            showlegend=False
        ))

        fig.add_trace(go.Scatter(
            x=[self.rashomon_set.epsilon],
            y=[real_pattern_rashomon_ratio],
            mode='markers',
            marker=dict(color=contrast_colors[0], size=5),
            text = [f'Epsilon: {self.rashomon_set.epsilon:.2f} <br> Pattern Rashomon Ratio : {real_pattern_rashomon_ratio:.2f}'],
            hoverinfo="text",
            name = "Pattern Rashomon Ratio for the given Rashomon Set."
        ))

        step = max(1, len(valid_epsilon)//10)
        tickvals = [round(e, 2) for e in valid_epsilon[::step]]

        tickfont_combined = {**plot_axis_font, "size": 9}
        fig.update_layout(
            title = dict(text = 'Pattern Rashomon Ratio vs Epsilon values',
                        font = plot_title_font,
                        x=0.5, 
                        xanchor='center'),
            height=400, 
            width = 600,
            xaxis = dict(title=dict(
                            text="Epsilon",
                            font=plot_axis_font),
                         showgrid=False, 
                         tickmode='array', 
                         tickvals = tickvals, 
                         tickangle=45, 
                         tickfont=tickfont_combined),
            yaxis=dict(title=dict(
                            text="Pattern Rashomon Ratio",
                            font=plot_axis_font),
                       range = [0,1.05], 
                       showgrid=False, 
                       tickfont=tickfont_combined),
            margin= dict(l=70, r = 40, t = 60, b = 60),
            legend=dict(
                font=dict(size=9),
                x=1,             
                y=0,             
                xanchor='right',
                yanchor='bottom',
                orientation='v'
            ))

        descr= """
                The plot shows how Pattern Rashomon Ratio changes with epsilon values. The highlighted point is the Rashomon Ratio calculated for the given Rashomon Set. 
            """
        
        return fig, descr
    
    #override
    def generate_rashomon_set_table(self)-> Tuple[pd.DataFrame, str]:
        '''
            Method for generating the table with Rashomon Intersection models and their base metrics values, as well as a brief text overview of the Rashomon Intersection properties.
            
            Returns:
                Tuple[go.Figure, str]
                - plot
                - plot description
        '''
        leaderboard= self.rashomon_set.leaderboard
        rashomon_set = self.rashomon_set.rashomon_set
        base_model_name = self.rashomon_set.base_model

        filtered_leaderboard = leaderboard.loc[leaderboard['model'].isin(rashomon_set), ['model']+ self.rashomon_set.metrics].reset_index(drop=True).round({metric : 3 for metric in self.rashomon_set.metrics})
        base_row = filtered_leaderboard[filtered_leaderboard['model'] == base_model_name]
        other_rows = filtered_leaderboard[filtered_leaderboard['model'] != base_model_name]
        filtered_leaderboard = pd.concat([base_row, other_rows], ignore_index=True)
        
        filtered_leaderboard.insert(0, "Nr", range(1, len(filtered_leaderboard)+1))
        styled_df = filtered_leaderboard.copy().astype(str)
        
        mask = styled_df["model"] == base_model_name
        styled_df.loc[mask] = styled_df.loc[mask].apply(
            lambda col: col.map(lambda x: f"<b>{x}</b>")
        )

        fig = go.Figure(data = [go.Table(
            columnwidth = [30,200,100],
            header = dict(
                values=list(filtered_leaderboard.columns),
                font=dict(family="Inter, sans-serif", color='darkgray', size=13),
                fill_color='white',
                align='left'
            ),
            cells = dict(values = [styled_df[c] for c in styled_df.columns],
                         align = 'left', fill_color='white', font=dict(family="Inter, sans-serif",color='darkgray', size=12))
        )])

        le, names = self.rashomon_set.find_same_score_as_base()
        if le == 1:
            descr =f"""
                The table shows all models that are included in the Rashomon Intersection for the metrics  : {self.rashomon_set.metrics[0]}, {self.rashomon_set.metrics[1]} and epsilon value : {self.rashomon_set.epsilon:.3f}. The values of the base metrics are rounded to three decimal points. 
                <br>The intersection consists of <strong>{len(self.rashomon_set.rashomon_set)} models </strong>  with <strong>{self.rashomon_set.base_model} </strong>as the base model.
                <br>There is <strong>{le} model</strong>, which achieved the same value of the metrics weighted sum as the base model - {names[0]}
                """
        elif le == 0:
            descr =f"""
                The table shows all models that are included in the Rashomon Intersection for the metrics : {self.rashomon_set.metrics[0]}, {self.rashomon_set.metrics[1]} and epsilon value : {self.rashomon_set.epsilon:.3f}. The values of the base metrics are rounded to three decimal points. 
                <br>The intersection consists of <strong>{len(self.rashomon_set.rashomon_set)} models </strong>  with <strong>{self.rashomon_set.base_model} </strong>as the base model.
                <br>There are <strong>no models</strong>, which achieved the same value of the metrics weighted sum as the base model.
                """
        else:
            descr =f"""
                The table shows all models that are included in the Rashomon Intersection for the metrics : {self.rashomon_set.metrics[0]}, {self.rashomon_set.metrics[1]} and epsilon value : {self.rashomon_set.epsilon:.3f}. The values of the base metrics are rounded to three decimal points. 
                <br>The set consists of <strong>{len(self.rashomon_set.rashomon_set)} models </strong>  with <strong>{self.rashomon_set.base_model} </strong>as the base model.
                <br>There are <strong>{le} models</strong>, which achieved the same value of the metrics weighted sum as the base model - {', '.join(f"{x}" for x in names)}
                """

        return fig, descr
    
    def generate_gauge_for_metrics(self)-> Tuple[go.Figure, str]:
        '''
            Method for generating gauge plots that visualize the proportion of models from each metric-specific Rashomon set that are included in the Rashomon Intersection.

            Returns:
                    Tuple[go.Figure, str]
                    - plot
                    - plot description
        '''

        intersection_metrics = self.rashomon_set.metrics
        intersection_models = self.rashomon_set.rashomon_set
        rashomon_models_metric1 = self.rashomon_set.get_rashomon_set_for_metric(intersection_metrics[0])
        rashomon_models_metric2 = self.rashomon_set.get_rashomon_set_for_metric(intersection_metrics[1])

        rashomon_size_metric1, rashomon_size_metric2 = len(rashomon_models_metric1), len(rashomon_models_metric2)
        models_in_intersection_metric1 = len(set(rashomon_models_metric1) & set(intersection_models))
        models_in_intersection_metric2 = len(set(rashomon_models_metric2) & set(intersection_models))


        fig = make_subplots(
            rows=1, cols=2,
            specs=[[{"type": "indicator"}, {"type": "indicator"}]]
        )

        fig.add_trace(go.Indicator(
            mode='number+gauge',
            value=models_in_intersection_metric1,
            title={"text": f"{intersection_metrics[0]}", "font": plot_title_font},
            number={'font': {'size': 30, 'color': '#6B7F8A'}},
            gauge={
                "axis": {"range": [0, rashomon_size_metric1]},
                "bar": {"color": "#6C929E"},
                "shape": "angular"
            }
        ), row=1, col=1)

        fig.add_trace(go.Indicator(
            mode='number+gauge',
            value=models_in_intersection_metric2,
            title={"text": f"{intersection_metrics[1]}", "font" : plot_title_font},
            number={'font': {'size': 30, 'color': '#6B7F8A'}},
            gauge={
                "axis": {"range": [0, rashomon_size_metric2]},
                "bar": {"color": "#6C929E"},
                "shape": "angular"}
        ), row=1, col=2)
        

        fig.update_layout(
            autosize=True,
            margin=dict(l=50, r=50, t=0, b=0)
        )

        descr = f'''Gauge plots showing how many models from each metric-specific Rashomon set 
        ({intersection_metrics[0]} and {intersection_metrics[1]}) were included in the final Rashomon Intersection.
        The left gauge shows {models_in_intersection_metric1} out of {rashomon_size_metric1} models from {intersection_metrics[0]} Rashomon set,
        and the right gauge shows {models_in_intersection_metric2} out of {rashomon_size_metric2} models from {intersection_metrics[1]} Rashomon set.
        '''

        return fig, descr


    def plot_venn_diagram(self)-> Tuple[go.Figure, str]:
        '''
            Method for visualizing rashomon sets for both metrics as well as the intersection using the Venn diagram. 

            Returns:
                Tuple[go.Figure, str]
                - plot
                - plot description
        '''

        #obtaining sets for both metrics and the intersection
        metrics = self.rashomon_set.metrics
        metric1_set = self.rashomon_set.get_rashomon_set_for_metric(metrics[0])
        metric2_set = self.rashomon_set.get_rashomon_set_for_metric(metrics[1])
        metric1_set = set(metric1_set)
        metric2_set = set(metric2_set)
        intersection = self.rashomon_set.rashomon_set
        set_A = metric1_set - metric2_set
        set_B = metric2_set - metric1_set
        set_AB = set(intersection)
        set_A_str = [str(item) for item in set_A]
        set_B_str = [str(item) for item in set_B]
        set_AB_str = [str(item) for item in set_AB]
        
        circle_radius = 2
        circle_A = {'x': -1, 'y': 0, 'r': 2, 'color': 'rgba(108, 146, 158, 0.3)' }

        circle_B = {'x': 1, 'y': 0, 'r': 2, 'color': 'rgba(85, 129, 177, 0.3)'}
        
        fig = go.Figure()
    
        for circle in [circle_A, circle_B]:
            fig.add_shape(
                type="circle",
                xref="x", yref="y",
                x0=circle['x']-circle['r'], y0=circle['y']-circle['r'],
                x1=circle['x']+circle['r'], y1=circle['y']+circle['r'],
                line=dict(color=circle['color'].replace('0.3', '1'), width=2),
                fillcolor=circle['color']
            )
        
        fig.add_trace(go.Scatter(
            x=[circle_A['x']], y=[circle_A['y'] + circle_radius + 0.3],
            mode="text", text=[f"<b>{metrics[0]}</b>"],
            textposition="middle center",
            textfont=dict(size=14, color='rgba(108, 146, 158, 1)'),
            showlegend=False
        ))
        
        fig.add_trace(go.Scatter(
            x=[circle_B['x']], y=[circle_B['y'] + circle_radius + 0.3],
            mode="text", text=[f"<b>{metrics[1]}</b>"],
            textposition="middle center",
            textfont=dict(size=14, color= 'rgba(85, 129, 177, 1)'),
            showlegend=False
        ))
        
        def add_text_list(x_center, y_center, items, position="left", max_display=15):
            if not items:
                return
            if len(self.rashomon_set.leaderboard)>12:
                font_size = 7.5
            else:
                font_size=12
                
            display_items = items[:max_display]
            text = '<br>'.join(display_items)
            if len(items) > max_display:
                text += f'<br>... +{len(items)-max_display} more'
            
            
            if position == "left":
                x_pos = x_center - circle_radius * 0.38
            elif position == "right":
                x_pos = x_center + circle_radius * 0.38 
            else:  # "center"
                x_pos = x_center
                
            fig.add_trace(go.Scatter(
                x=[x_pos], y=[y_center],
                mode="text",
                text=[text],
                textposition="middle center",
                textfont=dict(size=font_size, color='black'),
                showlegend=False
            ))
        
        add_text_list(circle_A['x'], circle_A['y'], set_A_str, position="left")    
        add_text_list(circle_B['x'], circle_B['y'], set_B_str, position="right")   
        add_text_list(0, 0, set_AB_str, position="center")                         
        
        x_range = [-2.5, 2.5]
        y_range = [-2, 2.5]
        
        fig.update_layout(
            xaxis=dict(visible=False, range=x_range),
            yaxis=dict(visible=False, range=y_range),
            plot_bgcolor='white',
            showlegend=False,
            margin=dict(l=0, r=0, b=5, t=40),
            autosize=True,
        )
        
        fig.add_annotation(
            x=0, y=-2.2,
            text=f"{metrics[0]}: {len(set_A)} | {metrics[1]}: {len(set_B)} | Intersection: {len(set_AB)}",
            showarrow=False,
            font=dict(size=12, color='black')
        )

        descr = "The Venn diagram represents models that are included in the separate Rashomon Sets for the given parameters as well as in the Rashomon Intersection."

        return fig, descr
    

    def pareto_front_plot(self)->Tuple[go.Figure, str]:
        '''
            Method for visualizing the Pareto Front created from all models available in the leaderboard as an additional analysis to the RashomonIntersection.
            
            Returns:
                Tuple[go.Figure, str]
                - plot
                - plot description
        '''

        def pareto_front(df, x_col, y_col):
            #data = df.sort_values(by=[x_col, y_col], ascending=[False, False]).reset_index(drop=True)
            pareto_indices =[]
            

            for i in df.index.values: #point that we check
                isDominant = True
                xi = df.loc[i, x_col]
                yi = df.loc[i, y_col]

                for j in df.index.values: #all other points to check dominance
                    if i==j:
                        continue
                    xj = df.loc[j, x_col]
                    yj = df.loc[j, y_col]
                    condition1 = (xj>=xi) and (yj>=yi) #xj is not worse in any metric
                    condition2 = (xj>xi) or (yj>yi) # j is better in at least one metric
                    if condition1 and condition2:
                        isDominant = False #there is a better point j 
                        break
                if isDominant:
                    pareto_indices.append(i)
            return df.loc[pareto_indices]
        

        whole_leaderboard = self.rashomon_set.leaderboard
        
        metric1 = self.rashomon_set.metrics[0]
        metric2 = self.rashomon_set.metrics[1]
        
        agg = whole_leaderboard.groupby([metric1, metric2]).agg(
            count=('model','size'),
            models=('model', lambda x: ', '.join(x))
        ).reset_index()

        front = pareto_front(agg, metric1, metric2)
        
        agg['color'] = chart_colors_v2[6]
        agg.loc[agg.set_index([metric1, metric2]).index.isin(front.set_index([metric1, metric2]).index), 'color'] = contrast_colors[0]
        agg['hover_text'] = agg.apply(lambda row: 
                              f"Number of models: {row['count']}<br>" +
                              "<br>".join([f"- {m}" for m in row['models'].split(', ')] ) +
                              f"<br>{metric1}: {row[metric1]}<br>{metric2}: {row[metric2]}", axis=1)

        normal_points = agg[agg['color'] ==chart_colors_v2[6] ]
        pareto_points = agg[agg['color'] ==contrast_colors[0] ]

        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x = normal_points[metric1],
            y = normal_points[metric2],
            mode = 'markers',
            marker=dict(
                size=normal_points['count'] , 
                color=chart_colors_v2[6],
                sizemode = "area",
                sizeref = 20 * max(agg['count']) / (40 ** 2),
                sizemin=3
            ),
            text=normal_points['hover_text'],
            hoverinfo='text',
            name = 'Other models'

                ))
        
        fig.add_trace(go.Scatter(
            x = pareto_points[metric1],
            y = pareto_points[metric2],
            mode = 'markers',
            marker=dict(
                size=pareto_points['count'],  
                color=contrast_colors[0] ,
                sizemode = "area",
                sizeref = 20 * max(agg['count']) / (40 ** 2),
                sizemin=3
            ),
            text=pareto_points['hover_text'],
            hoverinfo='text',
            name = 'Pareto Front'

                ))
        
        
        tickfont_combined = {**plot_axis_font, "size": 9}
        fig.update_layout(
            title = dict(text = 'All models from the leaderboard',
                        font = plot_title_font),
            xaxis = dict(title=dict(
                            text=f"{metric1}",
                            font=plot_axis_font),
                         tickfont=tickfont_combined),
            yaxis=dict(title=dict(
                            text=f"{metric2}",
                            font=plot_axis_font),
                       tickfont=tickfont_combined,
                       gridcolor = "#D3D3D3"),
            legend=dict(
                font=plot_axis_font,
                itemsizing='constant'
            )
        )

        

        pareto_leaderboard = pareto_front(whole_leaderboard, metric1, metric2)
        pareto_models = pareto_leaderboard['model'].tolist()

        return fig, f"The plot illustrates all models present in the <strong>leaderboard</strong> and their scores for both evaluation metrics. The highlighted points belong to the Pareto Front, meaning that for each of these models, there is no other model that simultaneously achieves better scores in all evaluation metrics. At least one metric is superior compared to other models.<br> Models in the Pareto Front: <br>  {', '.join(pareto_models)}"


from abc import ABC, abstractmethod

class Builder(ABC):

    '''
        Abstract class asserting that all pipeline classes have implemented methods such as:
        preview_rashomon(), set_epsilon() and build()
    '''

    @abstractmethod
    def preview_rashomon(self):
        '''
            Method for printing all trained models and their scores (leaderboard) and the Rashomon Set size plot
            to help the user choose the right epsilon parameter value.
        '''
        pass

    @abstractmethod
    def set_epsilon(self):
        '''
            setting epsilon parameter value. Needs to be completed before the build() method.
        '''
        pass

    @abstractmethod
    def build(self):
        '''
            Core method for performing all pipeline steps. Has to assert that all parameters including epsilon have been specified.
        '''
        pass

    @abstractmethod
    def dashboard_close(self):
        '''
            Method closing all stramlit processes.
        '''
        pass